# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from .common import Writer, SimulatedWriter, WriterOptions
from ..extractors.common import Extractor
from ..exceptions import WriterNotFoundError
from ..session import SessionOptions

# from .strictyaml import StrictYamlWriter
import inspect

modules = ["babyl", "maildir", "mbox", "mh", "mmdf", "jsonl"]


def find(
    extractor: Extractor,
    module_name: str,
    session_options: SessionOptions,
    writer_options: WriterOptions,
):
    if session_options.get_urls:
        return SimulatedWriter(extractor, writer_options)

    globals_ = globals()

    if module_name in modules:
        module = __import__(module_name, globals_, None, (), 1)

        for cls in module.__dict__.values():
            if (
                inspect.isclass(cls)
                and not inspect.isabstract(cls)
                and issubclass(cls, Writer)
            ):
                return cls(extractor, writer_options)

    raise WriterNotFoundError(module_name)



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from mailbox import MH, MHMessage

from .common import FolderedMailWriter, WriterOptions
from ..extractors.common import Extractor


class MhWriter(FolderedMailWriter):
    tests = []

    def __init__(self, extractor: Extractor, options: WriterOptions):
        super().__init__(
            extractor,
            MH(options.output_path),
            options,
        )

    def _new_message(self):
        return MHMessage()



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse, urlunparse
from dataclasses import dataclass
from itertools import islice
import bs4
import re

from .common import normalize_url, regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


@dataclass  # (kw_only=True)
class HypermailPageState(PageState):
    relative_urls: list[str]


class HypermailExtractor(Extractor):
    tests = [
        {
            "url": "https://hypermail-project.org/archive/08/index.html",
            "test_base_url": "https://hypermail-project.org/archive/",
            "test_titles_hash": "83b453b64d8d916717aebda6528f5371d50a3c57",
            "test_item_count": 1155,
        },
        {
            "url": "https://hypermail-project.org/archive/98/0001.html",
            "test_base_url": "https://hypermail-project.org/archive/",
            "test_contents_hash": "798dc3a69b3be638373447ad6f8ccd287297d03a",
            "test_item_count": 15,
        },
    ]

    _page_href_regex = re.compile(r"^(\d+)/index.html$")
    _post_href_regex = re.compile(r"^(\d+).html$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(
            normalize_url(url, remove_suffixes=[], append_slash=False)
        )
        soup = Soup(response.content)

        _ = soup.find(
            "meta",
            attrs={"name": "generator", "content": re.compile("^hypermail.*$")},
        )

        header_metas = soup.try_find(
            "meta", attrs={"name": re.compile("^(Author)|(Subject)|(Date)$")}
        )
        title_title = soup.try_find(
            "title",
            string=re.compile(
                "^.*?(by thread)|(by author)|(with attachments)|(by date)$"
            ),
        )

        if header_metas or title_title:
            parsed_url = urlparse(response.url)
            path = PurePosixPath(parsed_url.path)
            base_url = normalize_url(
                str(
                    urlunparse(
                        parsed_url._replace(path=str(PurePosixPath(*path.parts[:-2])))
                    )
                )
            )
            return HypermailExtractor(session, base_url, options)

        return HypermailExtractor(session, response.url, options)

    def _fetch_top_boards(self):
        pass

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        resolved_url = normalize_url(response.url, append_slash=False)

        if resolved_url == self.base_url:
            return self.root

        parsed_url = urlparse(resolved_url)
        path = PurePosixPath(parsed_url.path)

        if len(path.parts) >= 2 and self._post_href_regex.match(path.parts[-1]):
            id = path.parts[-1].removesuffix(".html")
            return Thread(
                path=(id,),
                url=url,
                origin=resolved_url,
                data={},
                title="",  # TODO.
            )

        return self.root

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if state.url == board.url:
            response = self._session.get(board.url)
            soup = Soup(response.content)

            page_anchors = soup.find_all("a", attrs={"href": self._page_href_regex})
            relative_urls = list(
                reversed([page_anchor.get("href") for page_anchor in page_anchors])
            )

            relative_url = relative_urls.pop()
            return HypermailPageState(
                url=urljoin(self.base_url, relative_url), relative_urls=relative_urls
            )

        state = cast(HypermailPageState, state)

        response = self._session.get(state.url)
        soup = Soup(response.content)

        messages_list_div = cast(
            bs4.element.Tag, soup.find("div", class_="messages-list")
        )
        root_ul = cast(bs4.element.Tag, messages_list_div.find("ul"))
        child_uls = root_ul.find_all("ul")

        for child_ul in child_uls:
            if not (
                thread_anchor := child_ul.try_find(
                    "a", attrs={"href": self._post_href_regex}
                )
            ):
                continue

            href = thread_anchor.get("href")
            id = regex_match(self._post_href_regex, href).group(1)
            yield Thread(
                path=(id,),
                url=urljoin(self.base_url, href),
                origin=response.url,
                data={},
                title="",  # TODO.
            )

        if state.relative_urls:
            relative_url = state.relative_urls.pop()
            return HypermailPageState(
                url=urljoin(self.base_url, relative_url),
                relative_urls=state.relative_urls,
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        if state.url == thread.url:
            state.url = urljoin(thread.url, ".")

        response = self._session.get(state.url)
        soup = Soup(response.content)

        root_anchor = soup.find("a", attrs={"href": f"{thread.path[-1]}.html"})
        root_pos = len(list(root_anchor.parents))

        href = root_anchor.get("href")
        yield self._fetch_post(state, thread.path, (), urljoin(thread.url, href))

        child_ul = root_anchor.find_next("ul")
        child_anchors = child_ul.find_all("a", attrs={"href": self._post_href_regex})

        prev_depth = 0
        subpath: list[str] = []

        for child_anchor in child_anchors:
            child_pos = len(list(child_anchor.parents))
            cur_depth = (child_pos - root_pos) // 2

            href = child_anchor.get("href")
            id = regex_match(self._post_href_regex, href).group(1)

            if cur_depth > prev_depth:
                subpath.append(id)
            else:
                subpath[-(prev_depth - cur_depth - 1) :] = [id]

            yield self._fetch_post(
                state, thread.path, tuple(subpath), urljoin(state.url, href)
            )

            prev_depth = cur_depth

    def _fetch_post(
        self,
        state: PageState,
        path: tuple[str, ...],
        subpath: tuple[str, ...],
        url: str,
    ):
        response = self._session.get(url)
        soup = Soup(response.content)

        author_meta = soup.find("meta", attrs={"name": "Author"})

        address = soup.find("address")

        return Post(
            path=path,
            subpath=subpath,
            url=url,
            origin=response.url,
            data={},
            author=author_meta.get("content"),
            content="".join(str(v) for v in islice(address.next_siblings, 1, None)),
        )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from abc import abstractmethod
from urllib.parse import urljoin, urlparse, parse_qs
import logging

from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session


class HackernewsExtractor(Extractor):
    tests = [
        {
            "url": "https://news.ycombinator.com",
            "test_base_url": "https://news.ycombinator.com/",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/item?id=1",
            "test_base_url": "https://news.ycombinator.com/",
            "test_contents_hash": "ead3fec1c434236fb09995727cb02c4b7e7dc3ca",
            "test_item_count": 18,
        },
        {
            "url": "https://news.ycombinator.com",
            "test_base_url": "https://news.ycombinator.com/",
            "initial_page": PageState(url="https://news.ycombinator.com/item?id=1000"),
            "item_count": 5,
            "test_titles_hash": "bd1ab966ced8daf8c68cc5ad9e51a7bd6c7b622c",
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/newest",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsNewExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/news",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsTopExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/best",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsBestExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/ask",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsAskExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/show",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsShowExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
        {
            "url": "https://news.ycombinator.com/jobs",
            "test_base_url": "https://news.ycombinator.com/",
            "test_extractor_type": "extractors.hackernews.HackernewsJobExtractor",
            "item_count": 5,
            "test_item_count": 5,
        },
    ]

    PAGE_SIZE = 1000

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        parsed_url = urlparse(url)
        if parsed_url.netloc == "news.ycombinator.com":
            if parsed_url.path == "/newest":
                return HackernewsNewExtractor(session, urljoin(url, "/"), options)

            if parsed_url.path == "/news":
                return HackernewsTopExtractor(session, urljoin(url, "/"), options)

            if parsed_url.path == "/best":
                return HackernewsBestExtractor(session, urljoin(url, "/"), options)

            if parsed_url.path == "/ask":
                return HackernewsAskExtractor(session, urljoin(url, "/"), options)

            if parsed_url.path == "/show":
                return HackernewsShowExtractor(session, urljoin(url, "/"), options)

            if parsed_url.path == "/jobs":
                return HackernewsJobExtractor(session, urljoin(url, "/"), options)

            return HackernewsExtractor(session, urljoin(url, "/"), options)

    def _calc_first_item_id(self, page_id: int):
        return 1 + (page_id * self.PAGE_SIZE)

    def _calc_page_id(self, item_id: int):
        return (item_id - 1) // self.PAGE_SIZE

    def _fetch_top_boards(self):
        firebase_url = f"https://hacker-news.firebaseio.com/v0/maxitem.json"
        self._max_item_id = int(self._session.get(firebase_url).content)
        self.pages: list[list[int]] = [[]] * (1 + self._calc_page_id(self._max_item_id))

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        parsed_url = urlparse(url)

        # The whole site.
        # TODO: Do this `if` properly by normalizing the URL beforehand.
        if parsed_url.path == "" or parsed_url.path == "/":
            return self.root
        # Thread.
        elif parsed_url.path == "/item":
            parsed_query = parse_qs(parsed_url.query)
            item_id = str(parsed_query["id"][0])

            return self._fetch_item_thread(int(item_id))

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _get_is_fetchable(self, item_id: int):
        if item_id > self._max_item_id:
            return False

        page_id = self._calc_page_id(item_id)

        if item_id in self.pages[page_id]:
            return False

        return True

    def _register_item(self, item_id: int):
        page_id = self._calc_page_id(item_id)

        if page_id >= len(self.pages):
            return False

        self.pages[page_id].append(item_id)
        return True

    def _fetch_item_thread(self, item_id: int):
        while True:
            firebase_url = f"https://hacker-news.firebaseio.com/v0/item/{item_id}.json"
            response = self._session.get(firebase_url)
            data = response.json()

            if "parent" in data:
                item_id = data["parent"]
            else:
                page_id = self._calc_page_id(item_id)
                self.pages[page_id].append(item_id)

                self._register_item(item_id)
                return Thread(
                    path=(
                        str(
                            item_id,
                        ),
                    ),
                    url=f"https://news.ycombinator.com/item?id={item_id}",
                    origin=response.url,
                    data=data,
                    title=data.get("title", None),
                )

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        # FIXME: This is ineffective, as we connect twice for each non-top item.
        # We make artificial pages of 1000 items.

        parsed_state_url = urlparse(state.url)
        parsed_query = parse_qs(parsed_state_url.query)

        if "id" in parsed_query:
            state_item_id = int(parsed_query["id"][0])
            page_id = self._calc_page_id(state_item_id)
        else:
            page_id = self._calc_page_id(self._max_item_id)

        # Remove pages above the state item id.
        del self.pages[page_id + 1 :]

        for item_id in reversed(
            range(
                self._calc_first_item_id(page_id), self._calc_first_item_id(page_id + 1)
            )
        ):
            if not self._get_is_fetchable(item_id):
                continue

            yield self._fetch_item_thread(item_id)

        if page_id > 0:
            new_state_item_id = self._calc_first_item_id(page_id) - 1
            return PageState(
                url=f"https://news.ycombinator.com/item?id={new_state_item_id}"
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        post_paths: list[tuple[str, ...]] = [()]

        i = 0
        while True:
            post_path = post_paths[i]

            if post_path:
                id = post_path[-1]
            else:
                id = thread.path[-1]

            firebase_url = f"https://hacker-news.firebaseio.com/v0/item/{id}.json"

            response = self._session.get(firebase_url)
            data = response.json()

            if data:
                self._register_item(int(id))
                yield Post(
                    path=thread.path,
                    subpath=post_path,
                    url=thread.url,
                    origin=response.url,
                    data=data,
                    author=data.get("by", ""),
                    content=data.get("text", ""),
                )

                for kid_id in data.get("kids", []):
                    post_paths.append(post_path + (str(kid_id),))

            else:
                logging.warning(f"Item at id={post_path[-1]} is null")

            i += 1
            if i == len(post_paths):
                break


class HackernewsSpecificExtractor(HackernewsExtractor):
    tests = []

    @staticmethod
    @abstractmethod
    def get_firebase_url() -> str:
        return ""

    def _get_node_from_url(self, url: str):
        return self.root

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        json = self._session.get(self.get_firebase_url()).json()

        for story_id in json:
            firebase_url = f"https://hacker-news.firebaseio.com/v0/item/{story_id}.json"
            response = self._session.get(firebase_url)
            data = response.json()

            yield Thread(
                path=(str(story_id),),
                url=f"https://news.ycombinator.com/item?id={story_id}",
                origin=response.url,
                data=data,
                title=data.get("title", ""),
            )


class HackernewsNewExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/newstories.json"


class HackernewsTopExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/topstories.json"


class HackernewsBestExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/beststories.json"


class HackernewsAskExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/askstories.json"


class HackernewsShowExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/showstories.json"


class HackernewsJobExtractor(HackernewsSpecificExtractor):
    @staticmethod
    def get_firebase_url():
        return "https://hacker-news.firebaseio.com/v0/jobstories.json"



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse, parse_qs
import re

from .common import get_relative_url, normalize_url, regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


class PhpbbExtractor(Extractor):
    tests = [
        {
            "url": "https://phpbb.com/community",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_boards": {
                ("47",): {
                    "title": "General",
                },
                ("551",): {
                    "title": "Support Forums",
                },
                ("451",): {
                    "title": "Extensions Forums",
                },
                ("471",): {
                    "title": "Styles Forums",
                },
                ("52",): {
                    "title": "Non-support Specific",
                },
                ("48",): {
                    "title": "phpBB Archives",
                },
            },
        },
        {
            "url": "https://phpbb.com/community/viewforum.php?f=556",
            "item_count": 0,
            "test_base_url": "https://www.phpbb.com/community/",
            "test_boards": {
                ("551", "556", "561"): {
                    "title": "[3.2.x] Convertors",
                },
                ("551", "556", "566"): {
                    "title": "[3.2.x] Translations",
                },
            },
        },
        {
            "url": "https://www.phpbb.com/community/viewtopic.php?t=2377611",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_contents_hash": "d98797411be739205db624165f24fc248c93157b",
            "test_item_count": 7,
        },
        {
            "url": "https://www.phpbb.com/community/viewtopic.php?t=2534156",
            "test_base_url": "https://www.phpbb.com/community/",
            "test_contents_hash": "d88e2da2e01e843f71c70f62a874542a416efc5d",
            "test_item_count": 1,
        },
    ]

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        # Check for the existence of "viewforum.php".
        response = session.try_get(
            urljoin(
                normalize_url(url, remove_suffixes=["viewforum.php", "viewtopic.php"]),
                "viewforum.php",
            )
        )

        # A rather crude way to detect: we look for a <html> tag with "dir" attribute.
        soup = Soup(response.text)
        soup.find("html", attrs={"dir": True})

        return PhpbbExtractor(
            session,
            normalize_url(
                response.url, remove_suffixes=["viewforum.php", "viewtopic.php"]
            ),
            options,
        )

    def _is_viewforum_url(self, url: str):
        parsed_url = urlparse(url)
        path = PurePosixPath(str(parsed_url.path))

        if len(path.parts) != 1:
            try:
                get_relative_url(url, self.base_url)
            except ValueError:
                return False

        if path.parts[-1] != "viewforum.php":
            return False

        parsed_query = parse_qs(parsed_url.query)

        if "f" not in parsed_query:
            return False

        return True

    def _is_viewforum_pagination_url(self, url: str):
        if not self._is_viewforum_url(url):
            return False

        parsed_url = urlparse(url)
        parsed_query = parse_qs(parsed_url.query)

        if not {"f", "start"} <= parsed_query.keys():
            return False

        return True

    def _is_viewtopic_url(self, url: str, thread_id: int | None = None):
        parsed_url = urlparse(url)
        path = PurePosixPath(str(parsed_url.path))

        if len(path.parts) != 1:
            try:
                get_relative_url(url, self.base_url)
            except ValueError:
                return False

        if path.parts[-1] != "viewtopic.php":
            return False

        parsed_query = parse_qs(parsed_url.query)

        if "t" not in parsed_query:
            return False

        return True

    def _is_viewtopic_pagination_url(self, url: str, thread_id: int | None = None):
        if not self._is_viewtopic_url(url, thread_id):
            return False

        parsed_url = urlparse(url)
        parsed_query = parse_qs(parsed_url.query)

        if not {"t", "start"} <= parsed_query.keys():
            return False

        return True

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        board_lis = soup.find_all("div", class_="forabg")

        for board_li in board_lis:
            header_li = board_li.find("li", class_="header")
            board_anchor = header_li.find("a")

            parsed_href = urlparse(board_anchor.get("href"))
            parsed_query = parse_qs(parsed_href.query)

            try:
                board_id = parsed_query["f"][0]
            except KeyError:
                continue

            board_title = board_anchor.string
            board_url = urljoin(self.base_url, f"viewforum.php?f={board_id}")

            self._set_board(
                path=(board_id,),
                url=board_url,
                origin=response.url,
                data={},
                title=board_title,
                are_subboards_fetched=True,
            )

            subboard_anchors = board_li.find_all("a", class_="forumtitle")

            for subboard_anchor in subboard_anchors:
                parsed_href = urlparse(subboard_anchor.get("href"))
                parsed_query = parse_qs(parsed_href.query)
                subboard_id = parsed_query["f"][0]
                subboard_title = subboard_anchor.string
                subboard_url = urljoin(self.base_url, f"viewforum.php?f={subboard_id}")

                self._set_board(
                    path=(
                        board_id,
                        subboard_id,
                    ),
                    url=subboard_url,
                    origin=response.url,
                    data={},
                    title=subboard_title,
                )

    def _fetch_subboards(self, board: Board):
        if board is self.root:
            return

        response = self._session.get(board.url)

        try:
            get_relative_url(response.url, self.base_url)
        except ValueError:
            return

        soup = Soup(response.content)

        subboard_anchors = soup.find_all("a", class_="forumtitle")

        for subboard_anchor in subboard_anchors:
            parsed_href = urlparse(subboard_anchor.get("href"))
            parsed_query = parse_qs(parsed_href.query)

            try:
                subboard_id = parsed_query["f"][0]
            except KeyError:
                continue

            subboard_title = subboard_anchor.string
            subboard_url = urljoin(self.base_url, f"viewforum.php?f={subboard_id}")

            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_url,
                origin=response.url,
                data={},
                title=subboard_title,
                are_subboards_fetched=True,
            )

    def _resolve_url(self, url: str):
        return normalize_url(self._session.get(url).url, keep_queries=["f", "t"])

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        resolved_url = normalize_url(response.url, keep_queries=["f", "t"])

        parsed_url = urlparse(resolved_url)
        parts = PurePosixPath(parsed_url.path).parts

        if parts[-1] == "viewforum.php":
            self._fetch_lower_boards(self.root)

            parsed_query = parse_qs(parsed_url.query)

            if "f" not in parsed_query:
                return self.root

            id = parsed_query["f"][0]

            for board in self._boards:
                if board is not self.root and board.path[-1] == id:
                    return board

            raise ValueError
        elif parts[-1] == "viewtopic.php":
            id = parse_qs(parsed_url.query)["t"][0]
            soup = Soup(response.content)
            breadcrumbs = soup.find(class_="breadcrumbs")

            breadcrumb_anchors = breadcrumbs.find_all("a", attrs={"itemprop": "item"})

            breadcrumb_urls = [
                self._resolve_url(urljoin(url, anchor.get("href")))
                for anchor in breadcrumb_anchors
            ]
            board = self.find_board_from_urls(tuple(breadcrumb_urls[1:]))

            title_h2 = soup.find("h2", class_="topic-title")
            title = title_h2.find("a").string

            return Thread(
                path=board.path + (id,),
                url=resolved_url,
                origin=resolved_url,
                data={},
                title=title,
            )
        elif normalize_url(resolved_url) == self.base_url:
            return self.root

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        parsed_url = urlparse(state.url)
        board_id = parse_qs(parsed_url.query)["f"][0]

        parsed_query = parse_qs(parsed_url.query)
        if "start" in parsed_query:
            cur_start = int(parsed_query["start"][0])
        else:
            cur_start = 0

        response = self._session.get(state.url)
        soup = Soup(response.content)
        topic_anchors = soup.find_all(
            "a", class_="topictitle", attrs={"href": self._is_viewtopic_url}
        )

        if not topic_anchors:
            topic_anchors = soup.find_all("a", attrs={"href": self._is_viewtopic_url})

        for topic_anchor in topic_anchors:
            href = urljoin(self.base_url, topic_anchor.get("href"))
            parsed_href = urlparse(href)
            parsed_query = parse_qs(parsed_href.query)
            thread_id = parsed_query["t"][0]

            yield Thread(
                path=board.path + (thread_id,),
                url=href,
                origin=response.url,
                data={},
                title=topic_anchor.string,
            )

        pagination_anchors = soup.find_all(
            "a", attrs={"href": self._is_viewforum_pagination_url}
        )

        # Look for pagination links. Always choose the one with the smallest increment.
        min_start = None

        for pagination_anchor in pagination_anchors:
            parsed_href = urlparse(pagination_anchor.get("href"))
            parsed_query = parse_qs(parsed_href.query)
            start = int(parsed_query["start"][0])

            if start > cur_start and (not min_start or start < min_start):
                min_start = start

        if min_start:
            return PageState(
                url=urljoin(
                    self.base_url, f"viewforum.php?f={board_id}&start={min_start}"
                )
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        parsed_url = urlparse(state.url)
        thread_id = parse_qs(parsed_url.query)["t"][0]

        parsed_query = parse_qs(parsed_url.query)
        if "start" in parsed_query:
            cur_start = int(parsed_query["start"][0])
        else:
            cur_start = 0

        response = self._session.get(state.url)
        soup = Soup(response.content)
        post_divs = soup.find_all("div", class_="post")

        for post_div in post_divs:
            id_div_regex = re.compile(r"^post_content(\d+)$")
            id_div = post_div.find("div", id=id_div_regex)
            content_div = post_div.find("div", class_="content")

            username_tag = post_div.find(
                {"a", "span"}, class_={"username", "username-coloured"}
            )

            url_h3 = post_div.find("h3")
            url_anchor = url_h3.find("a")

            yield Post(
                path=thread.path,
                subpath=(regex_match(id_div_regex, id_div.get("id")).group(1),),
                url=urljoin(response.url, url_anchor.get("href")),
                origin=response.url,
                data={},
                author=username_tag.string,
                content=str("".join(str(v) for v in content_div.contents)),
            )

        pagination_anchors = soup.find_all(
            "a",
            attrs={"href": self._is_viewtopic_pagination_url},
        )

        # Look for pagination links. Always choose the one with the smallest increment.
        min_start = None

        for pagination_anchor in pagination_anchors:
            parsed_href = urlparse(pagination_anchor.get("href"))
            parsed_query = parse_qs(parsed_href.query)
            start = int(parsed_query["start"][0])

            if start > cur_start and (not min_start or start < min_start):
                min_start = start

        if min_start:
            return PageState(
                url=urljoin(
                    self.base_url, f"viewtopic.php?t={thread_id}&start={min_start}"
                )
            )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

import re

from .common import (
    Extractor,
    ExtractorOptions,
    Board,
    Thread,
    Post,
    PageState,
    regex_match,
)
from ..session import Session
from ..soup import Soup


class InvisionExtractor(Extractor):
    tests = [
        {
            "url": "https://invisioncommunity.com/forums",
            "test_base_url": "https://invisioncommunity.com/forums/",
            "test_boards": {
                ("180", "528"): {
                    "title": "Invision Community Insider",
                },
                ("180", "499"): {
                    "title": "Feedback",
                },
                ("180", "320"): {
                    "title": "Community Manager Chat",
                },
                ("492", "505"): {
                    "title": "General Questions",
                },
                ("492", "497"): {
                    "title": "Technical Problems",
                },
                ("492", "497", "524"): {
                    "title": "Classic self-hosted technical help",
                },
                ("492", "500"): {
                    "title": "Design and Customization",
                },
                ("492", "477"): {
                    "title": "Community Manager Idea Sharing",
                },
                ("307", "504"): {
                    "title": "Developer Connection",
                },
                ("307", "521"): {
                    "title": "Marketplace",
                },
            },
        },
        {
            "url": "https://invisioncommunity.com/forums/topic/367687-important-seo-step-that-is-often-overlooked/",
            "test_base_url": "https://invisioncommunity.com/forums/",
            "test_contents_hash": "2a0a5ff4d936045e3148bd910f207e07ed4a0ed3",
            "test_item_count": 65,
        },
        {
            "url": "https://invisioncommunity.com/forums/topic/447328-guide-joels-guide-to-subscriptions/",
            "test_base_url": "https://invisioncommunity.com/forums/",
            "test_contents_hash": "1e0e0413cf82cf82bf83cfdf69813309594a06e4",
            "test_item_count": 52,
        },
    ]

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(url)
        soup = Soup(response.content)

        breadcrumbs_ul = soup.find("ul", attrs={"data-role": "breadcrumbList"})
        breadcrumb_lis = breadcrumbs_ul.find_all("li")
        base_url = url

        if len(breadcrumb_lis) >= 2:
            base_url = breadcrumb_lis[1].find("a").get("href")

        if soup.find("a", attrs={"title": "Invision Community"}):
            return InvisionExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        category_lis = soup.find_all("li", class_="cForumRow")
        for category_li in category_lis:
            category_id = category_li.get("data-categoryid")
            category_anchor = category_li.find("h2").find_all("a")[1]

            self._set_board(
                path=(category_id,),
                url=category_anchor.get("href"),
                origin=response.url,
                data={},
                title=category_anchor.string,
                are_subboards_fetched=True,
            )

            board_divs = category_li.find_all("div", class_="cForumGrid")
            for board_div in board_divs:
                board_id = board_div.get("data-forumid")
                board_h3 = board_div.find("h3", class_="cForumGrid__title")
                board_anchor = board_h3.find("a")

                self._set_board(
                    path=(category_id, board_id),
                    url=board_anchor.get("href"),
                    origin=response.url,
                    data={},
                    title=category_anchor.string,
                    are_subboards_fetched=True,
                )

        self._fetch_lower_boards(self.root)

    def _fetch_subboards(self, board: Board):
        if board is self.root:
            return

        response = self._session.get(board.url)
        soup = Soup(response.content)

        subboard_divs = soup.find_all("div", class_="cForumGrid")
        for subboard_div in subboard_divs:
            subboard_id = subboard_div.get("data-forumid")
            subboard_h3 = subboard_div.find("h3")
            subboard_anchor = subboard_h3.find("a")

            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_anchor.get("href"),
                origin=response.url,
                data={},
                title=subboard_anchor.string,
                are_subboards_fetched=True,
            )

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        soup = Soup(response.content)

        breadcrumbs_ul = soup.find("ul", attrs={"data-role": "breadcrumbList"})
        breadcrumb_lis = breadcrumbs_ul.find_all("li")

        if len(breadcrumb_lis) <= 2:
            return self.root

        # Thread.
        if soup.try_find("article"):
            board_href = breadcrumb_lis[-2].find("a").get("href")
            thread_id = soup.find("body").get("data-pageid")
            title_meta = soup.find("meta", attrs={"property": "og:title"})

            for cur_board in self._boards:
                if cur_board.url == board_href:
                    return Thread(
                        path=cur_board.path + (thread_id,),
                        url=url,
                        origin=response.url,
                        data={},
                        title=str(title_meta.get("content")),
                    )
        # Board.
        else:
            for cur_board in self._boards:
                if cur_board.url == url:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board is self.root:
            return None

        response = self._session.get(state.url)
        soup = Soup(response.content)

        thread_lis = soup.find_all(
            "li", attrs={"data-controller": "forums.frontforum.topicRow"}
        )
        for thread_li in thread_lis:
            thread_id = thread_li.get("data-rowid")
            title_h4 = thread_li.find("h4", class_="ipsDataItem_title")
            thread_anchor = title_h4.find("a", attrs={"title": True})

            yield Thread(
                path=board.path + (thread_id,),
                url=thread_anchor.get("href"),
                origin=response.url,
                data={},
                title=thread_anchor.get("title"),
            )

        next_page_link = soup.try_find("link", attrs={"rel": "next"})
        if next_page_link:
            return PageState(url=next_page_link.get("href"))

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        content_articles = soup.find_all("article", class_="ipsComment")

        for content_article in content_articles:
            content_div = content_article.find(
                "div", attrs={"data-role": "commentContent"}
            )
            author_div = content_article.find("div", class_="cAuthorPane_content")
            author_h3 = author_div.find("h3", class_="cAuthorPane_author")
            url_div = author_div.find("div")
            id = regex_match(
                re.compile(r"^elComment_(\d+)"), content_article.get("id")
            ).group(1)

            yield Post(
                path=thread.path,
                subpath=(id,),
                url=url_div.find("a").get("href"),
                origin=response.url,
                data={},
                author=author_h3.find("a").string,
                content="".join(str(v) for v in content_div.contents),
            )

        next_page_link = soup.try_find("link", attrs={"rel": "next"})
        if next_page_link:
            return PageState(url=next_page_link.get("href"))



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from mailbox import mbox, mboxMessage

from .common import MailWriter, WriterOptions
from ..extractors.common import Extractor


class MboxWriter(MailWriter):
    tests = []

    def __init__(self, extractor: Extractor, options: WriterOptions):
        super().__init__(extractor, mbox(options.output_path), options)

    def _new_message(self):
        return mboxMessage()



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from abc import ABC, abstractmethod
from dataclasses import dataclass
from urllib.parse import urlparse, urlunparse, parse_qs, urlencode
from pathlib import PurePosixPath

from ..session import Session
from ..exceptions import SearchError
from ..version import __version__


def get_relative_url(url: str, base_url: str):
    parsed_base_url = urlparse(base_url)
    parsed_url = urlparse(url)

    base_path = PurePosixPath(str(parsed_base_url.path))
    path = PurePosixPath(str(parsed_url.path))

    if str(base_path) == ".":
        return path

    return str(path.relative_to(base_path))


def normalize_url(
    url: str,
    remove_suffixes: list[str] = ["index.php"],
    append_slash: bool = True,
    keep_queries: list[str] = [],
):
    parsed_url = urlparse(url)
    new_path = parsed_url.path.removesuffix("/")

    if not keep_queries or not parsed_url.query:
        for remove_suffix in remove_suffixes:
            new_path = new_path.removesuffix(remove_suffix)

    new_path = new_path.removesuffix("/")

    query = parse_qs(parsed_url.query)
    new_query = {key: query[key] for key in keep_queries if key in query}

    new_parsed_url = parsed_url._replace(
        path=new_path, params="", query=urlencode(new_query, doseq=True), fragment=""
    )

    new_url = urlunparse(new_parsed_url)

    if append_slash and not new_parsed_url.query:
        return f"{new_url}/"

    return str(new_url)


def regex_match(pattern: Pattern[str], strings: list[str] | str):
    if isinstance(strings, str):
        strings = [strings]

    for string in strings:
        result = pattern.match(string)

        if result:
            return result

    raise ValueError


@dataclass
class ExtractorOptions:
    path: bool


@dataclass  # (kw_only=True)
class PageState:
    url: str


@dataclass
class Item:
    path: tuple[str, ...]
    url: str
    origin: str
    data: dict[str, Any]


@dataclass
class Post(Item):
    subpath: tuple[str, ...]
    author: str
    content: str


@dataclass
class Thread(Item):
    title: str


@dataclass
class Board(Item):
    title: str


class Extractor(ABC):
    tests: list[dict[str, Any]]

    @final
    @classmethod
    def detect(
        cls, session: Session, url: str, options: ExtractorOptions
    ) -> Extractor | None:
        try:
            return cls._detect(session, url, options)
        except SearchError:
            pass

    @staticmethod
    @abstractmethod
    def _detect(
        session: Session, url: str, options: ExtractorOptions
    ) -> Extractor | None:
        pass

    def __init__(self, session: Session, base_url: str, options: ExtractorOptions):
        self._session = session
        self.base_url = base_url
        self.root = Board(
            path=(), url=self._resolve_url(base_url), origin=base_url, data={}, title=""
        )
        self._boards: list[Board] = [self.root]
        self._subboards: dict[tuple[str, ...], dict[str, Board]] = {(): {}}
        self._are_subboards_fetched: dict[tuple[str, ...], bool] = {(): False}
        self._are_all_boards_fetched: bool = False
        self._options = options

        self.board_state: PageState | None = None
        self.thread_state: PageState | None = None

    @final
    def fetch(self):
        self._fetch_top_boards()
        # self._fetch_lower_boards(self.root)

    @abstractmethod
    def _fetch_top_boards(self):
        pass

    def _set_board(
        self,
        *,
        path: tuple[str, ...],
        replace_path: tuple[str, ...] | None = None,
        are_subboards_fetched: bool | None = None,
        **kwargs: Any,
    ):
        if replace_path is None:
            replace_path = path

        parent_board = self._find_board(replace_path[:-1])

        if replace_path[-1] in self._subboards[parent_board.path]:
            for k, v in kwargs.items():
                setattr(self._subboards[parent_board.path][replace_path[-1]], k, v)

            new_parent_board = self._find_board(path[:-1])
            self._subboards[new_parent_board.path][path[-1]] = self._subboards[
                parent_board.path
            ].pop(replace_path[-1])
            self._subboards[replace_path] = {}

            if are_subboards_fetched is not None:
                self._are_subboards_fetched[replace_path] = are_subboards_fetched
            else:
                self._are_subboards_fetched[replace_path] = False

            return self._subboards[new_parent_board.path][path[-1]]
        else:
            # We use self.root's type because it may be a subclass of Board.
            self._subboards[parent_board.path][replace_path[-1]] = type(self.root)(
                path=replace_path, **kwargs
            )
            self._subboards[replace_path] = {}
            self._boards.append(self._subboards[parent_board.path][replace_path[-1]])

            if are_subboards_fetched is not None:
                self._are_subboards_fetched[replace_path] = are_subboards_fetched
            else:
                self._are_subboards_fetched[replace_path] = False

            return self._subboards[parent_board.path][replace_path[-1]]

    @final
    def _fetch_lower_boards(self, board: Board):
        if self._are_all_boards_fetched:
            return

        i = 0
        while i < len(self._boards):
            cur_board = self._boards[i]

            if cur_board.path[: len(board.path)] == board.path:
                self._fetch_subboards(cur_board)

            i += 1

        if not board.path:
            self._are_all_boards_fetched = True

    @abstractmethod
    def _fetch_subboards(self, board: Board):
        pass

    def _resolve_url(self, url: str):
        return url

    @abstractmethod
    def _get_node_from_url(self, url: str) -> Item | tuple[str, ...]:
        pass

    @final
    def find_board(self, path: tuple[str, ...]):
        # For now, `find_board` always needs all subboards to be fetched.
        if not self._are_all_boards_fetched:
            self._fetch_lower_boards(self.root)

        return self._find_board(path)

    @final
    def _find_board(self, path: tuple[str, ...]):
        cur_board: Board = self.root

        for path_part in path:
            if path_part not in self._subboards[cur_board.path]:
                self._fetch_lazy_subboard(cur_board, path_part)

            cur_board = self._subboards[cur_board.path][path_part]

        return cur_board

    @final
    def find_board_from_urls(self, urls: tuple[str, ...]):
        cur_board: Board = self.root

        for url in urls:
            subboard_urls = [
                subboard.url for _, subboard in self._subboards[cur_board.path].items()
            ]

            if url not in subboard_urls:
                self._fetch_lazy_subboards(cur_board)

            for _, subboard in self._subboards[cur_board.path].items():
                if subboard.url == url:
                    cur_board = subboard

        return cur_board

    @final
    def node_from_url(self, url: str):
        node = self._get_node_from_url(self._resolve_url(url))

        if isinstance(node, Board):
            return self.find_board(node.path)

        return node

    def _fetch_lazy_subboard(self, board: Board, id: str) -> Board | None:
        if not self._are_subboards_fetched[board.path]:
            for _ in self._fetch_lazy_subboards(board):
                pass

            self._are_subboards_fetched[board.path] = True

        return self._subboards[board.path][id]

    @abstractmethod
    def _fetch_lazy_subboards(self, board: Board) -> Generator[Board, None, None]:
        self._fetch_subboards(board)
        return iter(self._subboards[board.path])

    @final
    def subboards(self, board: Board):
        if not self._are_subboards_fetched[board.path]:
            for _ in self._fetch_lazy_subboards(board):
                pass

            self._are_subboards_fetched[board.path] = True

        return self._subboards[board.path]

    @abstractmethod
    def _fetch_board_page_threads(
        self, board: Board, state: PageState
    ) -> Generator[Thread, None, PageState | None]:
        pass

    @final
    def _fetch_board_threads(
        self, board: Board, initial_state: PageState | None = None
    ):
        self.board_state = initial_state or PageState(url=board.url)
        while self.board_state:
            self.board_state = yield from self._fetch_board_page_threads(
                board, self.board_state
            )

    @abstractmethod
    def _fetch_thread_page_posts(
        self, thread: Thread, state: PageState
    ) -> Generator[Post, None, PageState | None]:
        pass

    @final
    def _fetch_thread_posts(
        self, thread: Thread, initial_state: PageState | None = None
    ):
        self.thread_state = initial_state or PageState(url=thread.url)
        while self.thread_state:
            self.thread_state = yield from self._fetch_thread_page_posts(
                thread, self.thread_state
            )

    @final
    def threads(self, board: Board, initial_state: PageState | None = None):
        yield from self._fetch_board_threads(board, initial_state)

    @final
    def posts(self, thread: Thread, initial_state: PageState | None = None):
        yield from self._fetch_thread_posts(thread, initial_state)



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from dataclasses import dataclass
from urllib.parse import urljoin
import re

from .common import normalize_url, regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


@dataclass  # (kw_only=True)
class SimplemachinesPageState(PageState):
    page: int


class SimplemachinesExtractor(Extractor):
    tests = [
        {
            "url": "https://simplemachines.org/community",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_boards": {
                ("2",): {
                    "title": "Simple Machines",
                },
                ("3",): {
                    "title": "SMF Support",
                },
                ("18",): {
                    "title": "Customizing SMF",
                },
                ("15",): {
                    "title": "SMF Development",
                },
                ("4",): {
                    "title": "General Community",
                },
                ("16",): {
                    "title": "Simple Machines Blogs",
                },
                ("5",): {
                    "title": "Archived Boards and Threads...",
                },
            },
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?board=255.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_boards": {
                ("3", "254", "255"): {
                    "title": "PostgreSQL Support",
                },
                ("3", "20", "134"): {
                    "title": "vBulletin",
                },
                ("4", "19"): {
                    "title": "Site Comments, Issues and Concerns",
                },
            },
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?topic=573.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            # "test_contents_hash": "9716f8e58a4bb20934ae026d5d1b7d0b5c9bb449",
            "test_item_count": 6,
        },
        {
            "url": "https://www.simplemachines.org/community/index.php?topic=581247.0",
            "test_base_url": "https://www.simplemachines.org/community/",
            "test_contents_hash": "bafb2382778b3717f490001548106f60b00b06ae",
            "test_item_count": 1,
        },
    ]

    _category_id_regex = re.compile(r"^c(\d+)$")
    _board_id_regex = re.compile(r"^b(\d+)$")
    _span_id_regex = re.compile(r"^msg_(\d+)$")
    _div_id_regex = re.compile(r"^msg_(\d+)$")
    _subject_id_regex = re.compile(r"^subject_(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(url)
        soup = Soup(response.content)

        link = soup.find("link", attrs={"rel": "contents"})
        base_url = normalize_url(link.get("href"))

        simplemachines_anchor = soup.find(
            "a",
            attrs={
                "href": "https://www.simplemachines.org",
                "title": "Simple Machines",
            },
        )

        if simplemachines_anchor:
            return SimplemachinesExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        category_anchors = soup.find_all("a", id=self._category_id_regex)
        for category_anchor in category_anchors:
            category_id = regex_match(
                self._category_id_regex, category_anchor.get("id")
            ).group(1)
            category_title = str(category_anchor.next_sibling).strip()

            self._set_board(
                path=(category_id,),
                url=urljoin(response.url, f"index.php#c{category_id}"),
                origin=response.url,
                data={},
                title=category_title,
                are_subboards_fetched=True,
            )

            for parent in category_anchor.parents:
                board_anchors = parent.find_all("a", id=self._board_id_regex)

                if board_anchors:
                    for board_anchor in board_anchors:
                        board_id = regex_match(
                            self._board_id_regex, board_anchor.get("id")
                        ).group(1)

                        self._set_board(
                            path=(category_id, board_id),
                            url=board_anchor.get("href"),
                            origin=response.url,
                            data={},
                            title=str(board_anchor.string).strip(),
                            are_subboards_fetched=True,
                        )
                    break

    def _fetch_subboards(self, board: Board):
        if not board.url:
            return

        # Don't fetch top boards.
        if len(board.path) <= 1:
            return

        response = self._session.get(board.url)
        soup = Soup(response.content)

        subboard_anchors = soup.find_all("a", attrs={"id": self._board_id_regex})

        for subboard_anchor in subboard_anchors:
            subboard_id = regex_match(
                self._board_id_regex, subboard_anchor.get("id")
            ).group(1)
            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_anchor.get("href"),
                origin=response.url,
                data={},
                title=str(subboard_anchor.string).strip(),
                are_subboards_fetched=True,
            )

    def _resolve_url(self, url: str):
        return normalize_url(
            self._session.get(url).url,
            append_slash=False,
            keep_queries=["board", "topic"],
        )

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        soup = Soup(response.content)

        breadcrumbs = soup.try_find(class_="navigate_section")
        if not breadcrumbs:
            breadcrumbs = soup.find(class_="linktree")

        breadcrumb_anchors = breadcrumbs.find_all("a")

        # Thread.
        if soup.try_find("div", id="forumposts"):
            breadcrumb_urls = [anchor.get("href") for anchor in breadcrumb_anchors]
            board = self.find_board_from_urls(tuple(breadcrumb_urls[1:-1]))

            topic_input = soup.find("input", attrs={"name": "topic"})
            thread_id = topic_input.get("value")
            title_title = soup.find("title")

            return Thread(
                path=board.path + (thread_id,),
                url=url,
                origin=response.url,
                data={},
                title=str(title_title.string),
            )

        # Board.
        else:
            self._fetch_lower_boards(self.root)

            board_href = self._resolve_url(breadcrumb_anchors[-1].get("href"))

            for cur_board in self._boards:
                if cur_board.url == board_href:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        if not state.url:
            return None

        if state.url == board.url:
            state = SimplemachinesPageState(url=state.url, page=1)

        state = cast(SimplemachinesPageState, state)

        response = self._session.get(state.url)
        soup = Soup(response.content)

        msg_spans = soup.find_all("span", id=self._span_id_regex)

        for msg_span in msg_spans:
            thread_id = regex_match(self._span_id_regex, msg_span.get("id")).group(1)
            msg_anchor = msg_span.tags[0]

            yield Thread(
                path=board.path + (thread_id,),
                url=msg_anchor.get("href"),
                origin=response.url,
                data={},
                title=str(msg_anchor.string),
            )

        next_page_anchor = soup.try_find(
            "a", class_="nav_page", string=str(state.page + 1)
        )
        if next_page_anchor:
            return SimplemachinesPageState(
                url=next_page_anchor.get("href"), page=state.page + 1
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        if state.url == thread.url:
            state = SimplemachinesPageState(url=state.url, page=1)

        state = cast(SimplemachinesPageState, state)

        response = self._session.get(state.url)
        soup = Soup(response.content)

        post_wrapper_divs = soup.find_all("div", class_="post_wrapper")

        for post_wrapper_div in post_wrapper_divs:
            msg_div = post_wrapper_div.find("div", id=self._div_id_regex)
            subject_div = post_wrapper_div.find("div", id=self._subject_id_regex)

            poster_div = post_wrapper_div.find("div", class_="poster")
            poster_h4 = poster_div.find("h4")

            yield Post(
                path=thread.path,
                subpath=(regex_match(self._div_id_regex, msg_div.get("id")).group(1),),
                url=subject_div.find("a").get("href"),
                origin=response.url,
                data={},
                author=str(poster_h4.find("a").string),
                content="".join(str(v) for v in msg_div.contents).strip(),
            )

        next_page_anchor = soup.try_find("a", class_="nav_page", string=str(state.page))
        if next_page_anchor:
            return SimplemachinesPageState(
                url=next_page_anchor.get("href"), page=state.page + 1
            )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore
from re import Pattern

from .exceptions import TagSearchError, AttributeSearchError, PropertyError
import bs4

SoupInput = Callable[[Any], bool] | Pattern[str] | set[str] | str | None


class Soup:
    def __init__(self, markup: str | bytes):
        self.soup = bs4.BeautifulSoup(markup, "lxml")

    def try_find(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        **kwargs: Any,
    ) -> SoupTag | None:
        result = self.soup.find(name, attrs, recursive, string, **kwargs)

        if result is not None and not isinstance(result, bs4.element.Tag):
            raise TagSearchError

        if result:
            return SoupTag(result)

    def find(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        **kwargs: Any,
    ) -> SoupTag:
        result = self.try_find(name, attrs, recursive, string, **kwargs)

        if not result:
            raise TagSearchError

        return result

    def find_all(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        limit: int | None = None,
        **kwargs: Any,
    ) -> list[SoupTag]:
        result = self.soup.find_all(name, attrs, recursive, string, limit, **kwargs)

        return [SoupTag(tag) for tag in result]


class SoupTag:
    def __init__(self, tag: bs4.element.Tag):
        self.tag = tag

    def try_find(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        **kwargs: Any,
    ) -> SoupTag | None:
        result = self.tag.find(name, attrs, recursive, string, **kwargs)

        if result is not None and not isinstance(result, bs4.element.Tag):
            raise TagSearchError

        if result:
            return SoupTag(result)

    def find(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        **kwargs: Any,
    ):
        result = self.try_find(name, attrs, recursive, string, **kwargs)

        if not result:
            raise TagSearchError

        return result

    def find_all(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        recursive: bool = True,
        string: SoupInput | None = None,
        limit: int | None = None,
        **kwargs: Any,
    ):
        result = self.tag.find_all(name, attrs, recursive, string, limit, **kwargs)

        return [SoupTag(tag) for tag in result]

    def find_next(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        string: SoupInput | None = None,
        **kwargs: Any,
    ):
        result = self.tag.find_next(name, attrs, string, **kwargs)

        if not isinstance(result, bs4.element.Tag):
            raise TagSearchError

        return SoupTag(result)

    def find_previous(
        self,
        name: SoupInput | None = None,
        attrs: dict[str, Any] = {},
        string: SoupInput | None = None,
        **kwargs: Any,
    ):
        result = self.tag.find_previous(name, attrs, string, **kwargs)

        if not isinstance(result, bs4.element.Tag):
            raise TagSearchError

        return SoupTag(result)

    def try_get(self, key: str, default: str | list[str] | None = None):
        return self.tag.get(key, default)

    def get(self, key: str, default: str | list[str] | None = None):
        result = self.try_get(key, default)

        if not isinstance(result, str):
            raise AttributeSearchError

        return result

    def get_list(self, key: str, default: str | list[str] | None = None) -> list[str]:
        result = self.tag.get(key, default)

        if not isinstance(result, list):
            raise AttributeSearchError

        return result

    def encode_contents(self):
        # TODO: Error handling?
        return self.tag.encode_contents()

    @property
    def string(self):
        if not self.tag.string:
            raise PropertyError

        return str(self.tag.string)

    @property
    def contents(self):
        if not self.tag.contents:
            raise PropertyError

        return self.tag.contents

    @property
    def tags(self):
        tags = [SoupTag(e) for e in self.contents if isinstance(e, bs4.element.Tag)]

        if not tags:
            raise PropertyError

        return tags

    @property
    def parents(self):
        return self.tag.parents

    @property
    def next_sibling(self):
        return self.tag.next_sibling

    @property
    def next_siblings(self):
        return self.tag.next_siblings

    @property
    def previous_sibling(self):
        return self.tag.previous_sibling

    @property
    def previous_siblings(self):
        return self.tag.previous_siblings



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse
import re

from .common import normalize_url
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


class HyperkittyExtractor(Extractor):
    tests = [
        {
            "url": "https://mail.python.org/archives/",
            "board_count": 5,
            "test_base_url": "https://mail.python.org/archives/",
            "test_boards": {
                ("numpy-discussion@python.org",): {
                    "title": "NumPy-Discussion",
                    # "content": "Discussion of Numerical Python",
                },
                ("scipy-dev@python.org",): {
                    "title": "SciPy-Dev",
                    # "content": "SciPy Developers List",
                },
                ("charlottepython@python.org",): {
                    "title": "CharlottePython",
                    # "content": "List for the Python user group in Charlotte, North Carolina.",
                },
                ("python-dev@python.org",): {
                    "title": "Python-Dev",
                    # "content": "Python core developers",
                },
                ("python-ideas@python.org",): {
                    "title": "Python-ideas",
                    # "content": "Discussions of speculative Python language ideas",
                },
            },
        },
        {
            "url": "https://mail.python.org/archives/list/mm3_test@python.org",
            "test_base_url": "https://mail.python.org/archives/",
            "test_titles_hash": "ecbc317707691b486caff90204ab15230fcc11a7",
            "test_item_count": 21,
        },
        {
            "url": "https://mail.python.org/archives/list/mm3_test@python.org/thread/NGFDAQJOJQDRPOU4WLRZEB55KAXPJWGN/",
            "test_base_url": "https://mail.python.org/archives/",
            "test_contents_hash": "a24cd22d0d460688f42b2bfa40805616075f22c0",
            "test_item_count": 5,
        },
        {
            "url": "https://mail.python.org/archives/list/mm3_test@python.org/thread/JNBVEQQBAI67DBT4HFXI3PO4APTKGQZO/",
            "test_base_url": "https://mail.python.org/archives/",
            "test_contents_hash": "ac64c15360e035c1bf5e4778869244ce81333844",
            "test_item_count": 2,
        },
        {
            "url": "https://mail.python.org/archives/list/mm3_test@python.org/thread/IWRIV5ULS4BHY43TRYCHE2TXJB7KZQ7U/",
            "test_base_url": "https://mail.python.org/archives/",
            "test_contents_hash": "887446d5b46aed3076cae2115a0ba408ba8133e0",
            "test_item_count": 1,
        },
    ]

    _reply_level_regex = re.compile(r"reply-level-(\d+)")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(normalize_url(url, append_slash=False))
        soup = Soup(response.content)

        if extractor := HyperkittyExtractor.detect_postorius(
            session, url, soup, options
        ):
            return extractor

        if extractor := HyperkittyExtractor.detect_hyperkitty(
            session, url, soup, options
        ):
            return extractor

    @staticmethod
    def detect_postorius(
        session: Session, url: str, soup: Soup, options: ExtractorOptions
    ):
        if not (footer := soup.try_find("footer")):
            return None

        if not footer.try_find("a", string="Postorius Documentation"):
            return False

        if not (nav_link_anchors := soup.find_all("a", class_="nav-link")):
            return None

        base_url = normalize_url(urljoin(url, nav_link_anchors[1].get("href")))
        return HyperkittyExtractor(session, base_url, options)

    @staticmethod
    def detect_hyperkitty(
        session: Session, url: str, soup: Soup, options: ExtractorOptions
    ):
        if not (footer := soup.try_find("footer")):
            return None

        if not footer.try_find("a", string="HyperKitty"):
            return False

        if not (navbar_brand_anchor := soup.try_find("a", class_="navbar-brand")):
            return None

        base_url = normalize_url(urljoin(url, navbar_brand_anchor.get("href")))
        return HyperkittyExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        pass

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        resolved_url = normalize_url(response.url)

        if resolved_url == self.base_url:
            return self.root

        parsed_url = urlparse(resolved_url)
        path = PurePosixPath(parsed_url.path)

        if len(path.parts) >= 4 and path.parts[-2] == "thread":
            board_id = path.parts[-3]
            thread_id = path.parts[-1]

            soup = Soup(response.content)
            thread_header_div = soup.find("div", class_="thread-header")
            thread_h3 = thread_header_div.find("h3")

            return Thread(
                path=(board_id, thread_id),
                url=resolved_url,
                origin=resolved_url,
                data={},
                title=thread_h3.string,
            )
        elif len(path.parts) >= 2 and path.parts[-2] == "list":
            return self.find_board((path.parts[-1],))

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        url = normalize_url(urljoin(self.base_url, f"list/{id}"))
        response = self._session.get(url)
        soup = Soup(response.content)

        title = ""

        if title_section := soup.try_find("section", id="title"):
            if h1 := title_section.try_find("h1"):
                title = h1.string.strip()
            elif h2 := title_section.try_find("h2"):
                title = h2.string.strip()

        # description = ""

        # if description_section := soup.find("p", id="description"):
        # description = description_section.string

        return self._set_board(
            path=(id,),
            url=url,
            origin=response.url,
            data={},
            # data={"description": description},
            title=title,
        )

    def _fetch_lazy_subboards(self, board: Board):
        href: str = ""
        url: str = self.base_url

        while href != "#":
            response = self._session.get(url)
            soup = Soup(response.content)
            list_anchors = soup.find_all("a", class_="list-name")

            for list_anchor in list_anchors:
                id = PurePosixPath(urlparse(list_anchor.get("href")).path).parts[-1]
                yield self._fetch_lazy_subboard(board, id)

            page_link_anchors = soup.find_all("a", class_="page-link")
            next_page_anchor = page_link_anchors[-1]

            href = next_page_anchor.get("href")
            url = urljoin(self.base_url, href)

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        match = re.match(r"^.*latest\?page=(\d+)$", state.url)

        if match:
            cur_page = int(match.group(1))
        else:
            cur_page = 1

        if board == self.root:
            return None

        if state.url == board.url:
            state.url = urljoin(state.url, "latest")

        origin = state.url

        response = self._session.get(origin)
        soup = Soup(response.content)

        thread_anchors = soup.find_all("a", class_="thread-title")

        if not thread_anchors:
            thread_spans = soup.find_all("span", class_="thread-title")
            thread_anchors = [thread_span.find("a") for thread_span in thread_spans]

        for thread_anchor in thread_anchors:
            yield Thread(
                path=board.path + (thread_anchor.get("name"),),
                url=urljoin(state.url, thread_anchor.get("href")),
                origin=origin,
                data={},
                title=str(thread_anchor.tag.contents[-1]).strip(),
            )

        if page_link_tags := soup.find_all(class_="page-link"):
            last_page = int(page_link_tags[-2].string)

            if cur_page < last_page:
                return PageState(url=urljoin(state.url, f"latest?page={cur_page + 1}"))

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        origin = state.url
        response = self._session.get(origin)

        if state.url == thread.url:
            soup = Soup(response.content)

            email_author_div = soup.find("div", class_="email-author")
            email_body_div = soup.find("div", class_="email-body")

            yield Post(
                path=thread.path,
                subpath=(),
                url=urljoin(
                    origin,
                    soup.find("div", class_="messagelink").find("a").get("href"),
                ),
                origin=origin,
                data={},
                author=str(email_author_div.find("a").string),
                content="".join(str(v) for v in email_body_div.contents),
            )

            return PageState(url=urljoin(state.url, "replies?sort=thread"))

        json = response.json()

        replies_html = json["replies_html"]
        soup = Soup(replies_html)

        reply_level_divs = soup.find_all("div", class_=["even", "odd"])
        prev_reply_level = 0
        subpath: list[str] = []

        for reply_level_div in reply_level_divs:
            for klass in reply_level_div.get_list("class"):
                if match := self._reply_level_regex.match(klass):
                    cur_reply_level = int(match.group(1))
                    break
            else:
                cur_reply_level = 0

            email_header_div = reply_level_div.find("div", class_="email-header")
            id = email_header_div.get("id")

            if cur_reply_level > prev_reply_level:
                subpath.append(id)
            else:
                subpath[-(prev_reply_level - cur_reply_level - 1) :] = [id]

            email_author_div = reply_level_div.find("div", class_="email-author")
            email_body_div = reply_level_div.find("div", class_="email-body")

            yield Post(
                path=thread.path,
                subpath=tuple(subpath),
                url=urljoin(
                    origin, soup.find("div", class_="messagelink").find("a").get("href")
                ),
                origin=origin,
                data={},
                author=str(email_author_div.find("a").string),
                content="".join(str(v) for v in email_body_div.contents),
            )

            prev_reply_level = cur_reply_level

        if json["more_pending"]:
            next_offset = json["next_offset"]
            return PageState(
                url=urljoin(state.url, f"replies?sort=thread&offset={next_offset}")
            )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse
import re

from .common import regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


class ProboardsExtractor(Extractor):
    tests = [
        {
            "url": "https://support.proboards.com",
            "test_base_url": "https://support.proboards.com/",
            "test_boards": {
                ("4",): {
                    "title": "Tech Support",
                },
                ("4", "80"): {
                    "title": "Help Guide",
                },
                ("4", "80", "138"): {
                    "title": "Help Guide Index",
                },
                ("4", "80", "76"): {
                    "title": "Admins and Moderators",
                },
                ("4", "80", "74"): {
                    "title": "Users and Members",
                },
                ("4", "80", "134"): {
                    "title": "Support FAQs",
                },
                ("4", "80", "72"): {
                    "title": "Plugin Developers",
                },
                ("4", "44"): {
                    "title": "Support Board",
                },
                ("4", "112"): {
                    "title": "Forums.net Sales Questions",
                },
                ("32",): {
                    "title": "ProBoards Development",
                },
                ("32", "92"): {
                    "title": "Development Blog",
                },
                ("32", "174"): {
                    "title": "ProBoards Contributions",
                },
                ("32", "176"): {
                    "title": "ProBoards v6 Beta",
                },
                ("16",): {
                    "title": "Coding, Development and Design",
                },
                ("16", "151"): {
                    "title": "Coding Help",
                },
                ("16", "38"): {
                    "title": "Plugins",
                },
                ("16", "38", "35"): {
                    "title": "Plugin Library",
                },
                ("16", "38", "36"): {
                    "title": "Request a Plugin",
                },
                ("16", "38", "96"): {
                    "title": "Plugin Development",
                },
                ("16", "82"): {
                    "title": "Themes",
                },
                ("16", "82", "64"): {
                    "title": "Theme Library",
                },
                ("16", "82", "86"): {
                    "title": "Request a Theme",
                },
                ("16", "45"): {
                    "title": "Templates",
                },
                ("16", "45", "78"): {
                    "title": "Template Library",
                },
                ("16", "45", "84"): {
                    "title": "Request a Template",
                },
                ("16", "55"): {
                    "title": "Headers & Footers",
                },
                ("16", "55", "90"): {
                    "title": "Headers & Footers Library",
                },
                ("16", "55", "88"): {
                    "title": "Request a Header & Footer Code",
                },
                ("16", "22"): {
                    "title": "Graphic Design",
                },
                # ("16", "22", "126"): {
                # "title": "Graphic Design Library",
                # },
                ("16", "22", "21"): {
                    "title": "Graphic Design Requests",
                },
                ("6",): {
                    "title": "Your Forum",
                },
                ("6", "140"): {
                    "title": "Open for Discussion - Staff of Forums Sharing Ideas",
                },
                ("6", "140", "40"): {
                    "title": "Promoting Your Forum Discussions",
                },
                ("6", "140", "40", "18"): {
                    "title": "Archived Threads",
                },
                ("6", "31"): {
                    "title": "Get Opinions About Your Forum",
                },
                ("6", "31", "179"): {
                    "title": "GOAYF Archive",
                },
                ("6", "1"): {
                    "title": "Advertise Your Forum",
                },
                ("6", "1", "2"): {
                    "title": "Art/Music/Tech",
                },
                ("6", "1", "4"): {
                    "title": "Graphics/Coding",
                },
                ("6", "1", "8"): {
                    "title": "ProBoards Related",
                },
                ("6", "1", "3"): {
                    "title": "General Forums",
                },
                ("6", "1", "9"): {
                    "title": "Roleplay Forums",
                },
                ("6", "1", "19"): {
                    "title": "Fan Forums",
                },
                ("6", "1", "32"): {
                    "title": "Archives",
                },
                ("6", "142"): {
                    "title": "Affiliate Exchange",
                },
                ("20",): {
                    "title": "General",
                },
                ("20", "33"): {
                    "title": "General Talk",
                },
                ("20", "33", "20"): {
                    "title": "The Game Board",
                },
                # ("20", "33", "20", "143"): {
                # "title": "Proposing A New Game",
                # },
                ("20", "33", "26"): {
                    "title": "Welcome",
                },
            },
        },
        {
            "url": "https://support.proboards.com/thread/426372/why-respond-ads-wants",
            "test_base_url": "https://support.proboards.com/",
            "test_contents_hash": "e55671633d9c91f614d441557806fabe9899d60e",
            "test_item_count": 76,
        },
        {
            "url": "https://support.proboards.com/thread/15052/visual-conceptions",
            "test_base_url": "https://support.proboards.com/",
            "test_contents_hash": "f655c7ddf487e7eb802c2123cef686f383433ee0",
            "test_item_count": 31,
        },
    ]

    _category_name_regex = re.compile(r"^category-(\d+)$")
    _board_id_regex = re.compile(r"^board-(\d+)$")
    _thread_class_regex = re.compile(r"^thread-(\d+)$")
    _post_id_regex = re.compile(r"^post-(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        parsed_url = urlparse(url)

        if parsed_url.netloc.endswith("proboards.com"):
            return ProboardsExtractor(session, urljoin(url, "/"), options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        category_anchors = soup.find_all("a", attrs={"name": self._category_name_regex})

        for category_anchor in category_anchors:
            category_id = regex_match(
                self._category_name_regex, category_anchor.get("name")
            ).group(1)

            title_div = category_anchor.find_next("div", class_="title_wrapper")

            self._set_board(
                path=(category_id,),
                url=urljoin(response.url, f"#{category_anchor.get('name')}"),
                origin=response.url,
                data={},
                title=title_div.string,
                are_subboards_fetched=True,
            )

            category_div = category_anchor.find_next("div", class_="boards")
            board_trs = category_div.find_all("tr", id=self._board_id_regex)

            for board_tr in board_trs:
                board_id = regex_match(self._board_id_regex, board_tr.get("id")).group(
                    1
                )
                board_anchor = board_tr.find("a", class_=self._board_id_regex)

                self._set_board(
                    path=(category_id, board_id),
                    url=urljoin(self.base_url, board_anchor.get("href")),
                    origin=response.url,
                    data={},
                    title=board_anchor.string,
                    are_subboards_fetched=True,
                )

        self._fetch_lower_boards(self.root)

    def _fetch_subboards(self, board: Board):
        if not board.url:
            return

        response = self._session.get(board.url)
        soup = Soup(response.content)

        subboard_trs = soup.find_all("tr", id=self._board_id_regex)
        for subboard_tr in subboard_trs:
            subboard_id = regex_match(
                self._board_id_regex, subboard_tr.get("id")
            ).group(1)
            subboard_anchor = subboard_tr.find("a", class_=self._board_id_regex)

            self._set_board(
                path=board.path + (subboard_id,),
                url=urljoin(self.base_url, subboard_anchor.get("href")),
                origin=response.url,
                data={},
                title=subboard_anchor.string,
                are_subboards_fetched=True,
            )

    def _get_node_from_url(self, url: str):
        parsed_url = urlparse(url)
        url_parts = PurePosixPath(parsed_url.path).parts

        if len(url_parts) <= 1:
            return self.root

        if url_parts[1] == "thread":
            response = self._session.get(url)
            soup = Soup(response.content)

            breadcrumbs_div = soup.find("div", class_="nav-tree-wrapper")
            breadcrumb_anchors = breadcrumbs_div.find_all(
                "a", attrs={"itemprop": "item"}
            )

            board_url = urljoin(self.base_url, breadcrumb_anchors[-2].get("href"))

            for cur_board in self._boards:
                if cur_board.url == board_url:
                    thread_link_anchor = soup.find("a", class_="thread-link")
                    thread_id = regex_match(
                        self._thread_class_regex,
                        thread_link_anchor.get_list("class"),
                    ).group(1)
                    title_h1 = soup.find("h1")

                    return Thread(
                        path=cur_board.path + (thread_id,),
                        url=url,
                        origin=response.url,
                        data={},
                        title=title_h1.string,
                    )
        elif url_parts[1] == "board":
            for cur_board in self._boards:
                if cur_board.path[-1] == url_parts[1]:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        response = self._session.get(state.url)
        soup = Soup(response.content)

        thread_anchors = soup.find_all("a", class_="thread-link")
        for thread_anchor in thread_anchors:
            thread_id = regex_match(
                self._thread_class_regex, thread_anchor.get_list("class")
            ).group(1)
            yield Thread(
                path=board.path + (thread_id,),
                url=urljoin(self.base_url, thread_anchor.get("href")),
                origin=response.url,
                data={},
                title=thread_anchor.string,
            )

        next_page_li = soup.try_find("li", class_="next")
        if not next_page_li:
            return None

        next_page_anchor = next_page_li.try_find("a")

        if next_page_anchor and next_page_anchor.try_get("href"):
            return PageState(
                url=urljoin(self.base_url, next_page_anchor.get("href")),
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        post_trs = soup.find_all("tr", class_="item")
        for post_tr in post_trs:
            user_anchor = post_tr.try_find("a", class_="o-user-link")
            message_div = post_tr.find("div", class_="message")
            id = regex_match(self._post_id_regex, post_tr.get("id")).group(1)

            yield Post(
                path=thread.path,
                subpath=(id,),
                url=urljoin(self.base_url, f"post/{id}/thread"),
                origin=response.url,
                data={},
                author=user_anchor.string if user_anchor else "",
                content=str("".join(str(v) for v in message_div.contents)),
            )

        next_page_li = soup.try_find("li", class_="next")
        if not next_page_li:
            return None

        next_page_anchor = next_page_li.try_find("a")

        if next_page_anchor and next_page_anchor.try_get("href"):
            return PageState(
                url=urljoin(self.base_url, next_page_anchor.get("href")),
            )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse, urlunparse
from dataclasses import dataclass
import bs4
import re

from .common import normalize_url, regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..exceptions import TagSearchError
from ..session import Session
from ..soup import Soup


@dataclass  # (kw_only=True)
class PipermailPageState(PageState):
    relative_urls: list[str]


class PipermailExtractor(Extractor):
    tests = [
        {
            "url": "http://lists.opensource.org/mailman/listinfo",
            "test_base_url": "https://lists.opensource.org/",
            "test_boards": {
                ("license-discuss@lists.opensource.org",): {
                    "title": "License-discuss",
                },
                ("license-review@lists.opensource.org",): {
                    "title": "License-review",
                },
                ("publicpolicy@lists.opensource.org",): {
                    "title": "Publicpolicy",
                },
            },
        },
        {
            "url": "http://lists.opensource.org/pipermail/osideas_lists.opensource.org/2020-May/thread.html",
            "test_base_url": "http://lists.opensource.org/",
            "test_titles_hash": "32b7035d3f25274b7c20b29b78c2991294fa72e7",
            "test_item_count": 3,
        },
        {
            "url": "http://lists.opensource.org/pipermail/license-review_lists.opensource.org/2008-January/000031.html",
            "test_base_url": "http://lists.opensource.org/",
            "test_item_count": 22,
        },
    ]

    _listinfo_href_regex = re.compile(r"^listinfo/(.+)$")
    _listinfo_title_regex = re.compile(r"^(.+) Info Page$")
    _pipermail_page_href_regex = re.compile(
        r"^\d\d\d\d-(January|February|March|April|May|June|July|August|September|October|November|December)/thread.html$"
    )
    _post_href_regex = re.compile(r"^(\d+).html$")
    _root_post_comment_regex = re.compile(r"^0 ([^-]+)- $")
    _child_post_comment_regex = re.compile(r"^(1|2|3) ([^-]+)-(.*?)-? $")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(url)
        resolved_url = normalize_url(response.url, append_slash=False)

        parsed_url = urlparse(resolved_url)
        path = PurePosixPath(parsed_url.path)

        if len(path.parts) >= 4 and path.parts[-4] == "pipermail":
            return PipermailExtractor(
                session,
                str(
                    urlunparse(
                        parsed_url._replace(path=str(PurePosixPath(*path.parts[:-4])))
                    )
                ),
                options,
            )
        elif len(path.parts) >= 3 and path.parts[-3] == "pipermail":
            return PipermailExtractor(
                session,
                str(
                    urlunparse(
                        parsed_url._replace(path=str(PurePosixPath(*path.parts[:-3])))
                    )
                ),
                options,
            )
        elif len(path.parts) >= 2 and (
            path.parts[-2] == "pipermail" or path.parts[-2] == "mailman"
        ):
            return PipermailExtractor(
                session,
                str(
                    urlunparse(
                        parsed_url._replace(path=str(PurePosixPath(*path.parts[:-2])))
                    )
                ),
                options,
            )
        elif len(path.parts) >= 1 and (
            path.parts[-1] == "pipermail" or path.parts[-1] == "mailman"
        ):
            return PipermailExtractor(
                session,
                str(
                    urlunparse(
                        parsed_url._replace(path=str(PurePosixPath(*path.parts[:-1])))
                    )
                ),
                options,
            )

    def _fetch_top_boards(self):
        pass

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        normalized_url = normalize_url(response.url)

        if normalized_url == self.base_url:
            return self.root

        parsed_url = urlparse(normalized_url)
        path = PurePosixPath(parsed_url.path)

        if len(path.parts) >= 4 and path.parts[-4] == "pipermail":
            if path.parts[-1] == "thread.html":
                return self.find_board((path.parts[-3],))

            id = path.parts[-1].removesuffix(".html")
            board_id = path.parts[-3]

            # TODO: Properly read board name instead.
            board_id.replace("_", "@")

            soup = Soup(response.content)
            title = soup.find("title").string

            return Thread(
                path=(board_id, id),
                url=url,
                origin=response.url,
                data={},
                title=title,
            )
        elif len(path.parts) >= 3 and path.parts[-3] == "pipermail":
            return self.find_board((path.parts[-2],))
        elif (
            len(path.parts) >= 3
            and path.parts[-3] == "mailman"
            and path.parts[-2] == "listinfo"
        ):
            return self.find_board((path.parts[-1],))
        elif len(path.parts) >= 2:
            if path.parts[-2] == "pipermail":
                return self.find_board((path.parts[-1],))

            return self.root

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        nice_id = id.replace("@", "_")

        url = normalize_url(urljoin(self.base_url, f"mailman/listinfo/{nice_id}"))
        response = self._session.get(url)
        soup = Soup(response.content)

        title_title = soup.find("title")
        title = regex_match(self._listinfo_title_regex, title_title.string).group(1)

        # body = str(soup.find_all("p")[2].contents[1])
        return self._set_board(
            path=(id,),
            url=url,
            origin=response.url,
            data={},
            # data={"body": body},
            title=title,
        )

    def _fetch_lazy_subboards(self, board: Board):
        # TODO use a for loop over _fetch_lazy_subboard() instead
        url = normalize_url(urljoin(self.base_url, f"mailman/listinfo"))
        response = self._session.get(url)
        soup = Soup(response.content)

        listinfo_anchors = soup.find_all("a", attrs={"href": self._listinfo_href_regex})

        for listinfo_anchor in listinfo_anchors:
            href = listinfo_anchor.get("href")
            id = regex_match(self._listinfo_href_regex, href).group(1)
            yield self._fetch_lazy_subboard(board, id)

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        if state.url == board.url:
            id = board.path[0]
            pipermail_url = urljoin(self.base_url, f"pipermail/{id}")

            response = self._session.get(pipermail_url)
            soup = Soup(response.content)

            page_anchors = soup.find_all(
                "a", attrs={"href": self._pipermail_page_href_regex}
            )

            relative_urls = list(
                reversed([page_anchor.get("href") for page_anchor in page_anchors])
            )

            relative_url = relative_urls.pop()
            return PipermailPageState(
                url=urljoin(urljoin(self.base_url, f"pipermail/{id}/"), relative_url),
                relative_urls=relative_urls,
            )

        state = cast(PipermailPageState, state)

        response = self._session.get(state.url)
        soup = Soup(response.content)

        root_comments = soup.soup.find_all(
            string=lambda text: isinstance(text, bs4.element.Comment)
            and bool(self._root_post_comment_regex.match(text))
        )

        for root_comment in root_comments:
            thread_anchor = root_comment.find_next(
                "a", attrs={"href": self._post_href_regex}
            )
            href = thread_anchor.get("href")
            id = regex_match(self._post_href_regex, href).group(1)

            yield Thread(
                path=board.path + (id,),
                url=urljoin(state.url, href),
                origin=response.url,
                data={},
                title=thread_anchor.string,
            )

        if state.relative_urls:
            relative_url = state.relative_urls.pop()
            board_id = board.path[0]
            return PipermailPageState(
                url=urljoin(
                    urljoin(self.base_url, f"pipermail/{board_id}/"), relative_url
                ),
                relative_urls=state.relative_urls,
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        if state.url == thread.url:
            state.url = urljoin(thread.url, "thread.html")

        response = self._session.get(state.url)
        soup = Soup(response.content)

        root_anchor = soup.find("a", attrs={"href": f"{thread.path[-1]}.html"})
        root_comment = root_anchor.tag.find_previous(
            string=lambda text: isinstance(text, bs4.element.Comment)
        )
        if not isinstance(root_comment, bs4.element.Comment):
            raise TagSearchError

        yield self._fetch_post(state, thread.path, (thread.path[-1],), thread.url)

        thread_long_id = regex_match(
            self._root_post_comment_regex, str(root_comment)
        ).group(1)

        child_comments = soup.soup.find_all(
            string=lambda text: isinstance(text, bs4.element.Comment)
            and bool(self._child_post_comment_regex.match(text))
            and (
                text.startswith(f"1 {thread_long_id}-")
                or text.startswith(f"2 {thread_long_id}-")
                or text.startswith(f"3 {thread_long_id}-")
            )
        )

        prev_long_ids = []
        subpath: list[str] = []

        for child_comment in child_comments:
            cur_long_ids = (
                regex_match(self._child_post_comment_regex, child_comment.string)
                .group(3)
                .split("-")
            )

            child_anchor = child_comment.find_next(
                "a", attrs={"href": self._post_href_regex}
            )
            href = child_anchor.get("href")
            id = regex_match(self._post_href_regex, href).group(1)

            if len(cur_long_ids) > len(prev_long_ids):
                subpath.append(id)
            else:
                subpath[-(len(prev_long_ids) - len(cur_long_ids) - 1) :] = [id]

            yield self._fetch_post(
                state, thread.path, tuple(subpath), urljoin(state.url, href)
            )

            prev_long_ids = cur_long_ids

    def _fetch_post(
        self,
        state: PageState,
        path: tuple[str, ...],
        subpath: tuple[str, ...],
        url: str,
    ):
        response = self._session.get(url)
        soup = Soup(response.content)

        content_pre = soup.find("pre")
        content = "".join(str(v) for v in content_pre.contents)
        content = re.sub(r"><i>(.*?\n)</i>", r">\1", content)

        author_b = soup.find("b")

        return Post(
            path=path,
            subpath=subpath,
            url=url,
            origin=response.url,
            data={},
            author=str(author_b.string),
            content=content,
        )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore
from types import ModuleType

import inspect

from .common import Extractor, ExtractorOptions
from ..exceptions import ExtractorNotFoundError
from ..session import Session, SessionOptions

modules = [
    "hackernews",
    "hyperkitty",
    "pipermail",
    "hypermail",
    "xenforo",
    "vbulletin",
    "proboards",
    "invision",
    "discourse",
    "simplemachines",
    "phpbb",
]


def find(
    url: str, session_options: SessionOptions, extractor_options: ExtractorOptions
):
    session = Session(session_options)

    for cls in list_classes():
        obj = cls.detect(session, url, extractor_options)
        if obj:
            return obj

    raise ExtractorNotFoundError(url)


def list_classes() -> Iterable[Any]:
    globals_ = globals()

    for module_name in modules:
        module = __import__(module_name, globals_, None, (), 1)
        yield from _get_classes(module)


def _get_classes(module: ModuleType):
    return [
        cls
        for cls in module.__dict__.values()
        if (
            inspect.isclass(cls)
            and not inspect.isabstract(cls)
            and issubclass(cls, Extractor)
        )
    ]



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from urllib.parse import quote_plus

from . import extractors
from . import writers
from .session import SessionOptions
from .extractors.common import ExtractorOptions
from .writers.common import WriterOptions


class ForumDl:
    def download(
        self,
        urls: list[str],
        output_format: str,
        session_options: SessionOptions,
        extractor_options: ExtractorOptions,
        writer_options: WriterOptions,
    ):
        for url in urls:
            self.download_url(
                url,
                output_format,
                session_options,
                extractor_options,
                writer_options,
            )

    def download_url(
        self,
        url: str,
        output_format: str,
        session_options: SessionOptions,
        extractor_options: ExtractorOptions,
        writer_options: WriterOptions,
    ):
        extractor = extractors.find(url, session_options, extractor_options)

        if extractor:
            extractor.fetch()
            writer_options.output_path = writer_options.output_path or quote_plus(url)
            writer = writers.find(
                extractor, output_format, session_options, writer_options
            )
            writer.write(url)

    def list_extractors(self):
        return extractors.modules

    def list_output_formats(self):
        return writers.modules



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from urllib.parse import urljoin
import re

from .common import regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


class VbulletinExtractor(Extractor):
    tests = [
        {
            "url": "https://forum.vbulletin.com",
            "test_base_url": "https://forum.vbulletin.com/",
            "test_boards": {
                ("14",): {
                    "title": "vBulletin Announcements",
                },
                ("14", "28"): {
                    "title": "vBulletin Announcements",
                },
                ("18",): {
                    "title": "vBulletin Sales and Feedback",
                },
                ("18", "40"): {
                    "title": "vBulletin Pre-sales Questions",
                },
                ("18", "29"): {
                    "title": "Site Feedback",
                },
                ("24",): {
                    "title": "vBulletin 5 Connect",
                },
                ("24", "72"): {
                    "title": "vBulletin 5 Connect Feedback",
                },
                ("24", "111"): {
                    "title": "Support Issues & Questions",
                },
                ("24", "111", "114"): {
                    "title": "vBulletin 5 Tutorials",
                },
                ("24", "70"): {
                    "title": "vBulletin 5 Suggestions",
                },
                ("24", "67"): {
                    "title": "vBulletin 5 Installs & Upgrades",
                },
                ("4014354",): {
                    "title": "vBCloud",
                },
                ("4014354", "4014355"): {
                    "title": "Account Management",
                },
                ("4014354", "4017486"): {
                    "title": "vB Cloud Support & Troubleshooting.",
                },
                ("4136723",): {
                    "title": "International Support",
                },
                ("4136723", "3954329"): {
                    "title": "Deutschsprachiger Support",
                },
                ("4136723", "3954329", "3967454"): {
                    "title": "vBulletin 5 - Fragen und Probleme",
                },
                ("4136723", "3954329", "3954332"): {
                    "title": "vBulletin 5 - Installation & Upgrade",
                },
                ("4136723", "3954329", "3954333"): {
                    "title": "vBulletin 5 - Tipps, Tricks & häufige Fragen",
                },
                ("4136723", "3954329", "3954334"): {
                    "title": "vBulletin 5 - Handbuch",
                },
                ("4136723", "4001685"): {
                    "title": "Assistance francophone",
                },
                ("4136723", "4055583"): {
                    "title": "Soporte de Español",
                },
                ("3955117",): {
                    "title": "vBulletin Mobile",
                },
                ("3955117", "62"): {
                    "title": "vBulletin Mobile Suite",
                },
                ("21",): {
                    "title": "vBulletin 4",
                },
                ("21", "55"): {
                    "title": "vB4 Support & Troubleshooting",
                },
                ("21", "55", "101"): {
                    "title": "vBulletin 4 Quick Tips and Customizations",
                },
                ("21", "54"): {
                    "title": "vBulletin 4 Installations and Upgrades",
                },
                ("20",): {
                    "title": "vBulletin 3.8",
                },
                ("20", "47"): {
                    "title": "vBulletin 3.8 Support & Troubleshooting",
                },
                ("20", "47", "97"): {
                    "title": "vBulletin Quick Tips and Customizations",
                },
                ("20", "48"): {
                    "title": "vBulletin 3.8 Installation and Upgrades",
                },
                ("13",): {
                    "title": "Customizing vBulletin",
                },
                ("13", "42"): {
                    "title": "vBulletin Languages & Phrases",
                },
                ("13", "42", "105"): {
                    "title": "vBulletin Mobile Languages",
                },
                ("13", "31"): {
                    "title": "vBulletin Templates, Graphics & Styles",
                },
                ("13", "45"): {
                    "title": "vBulletin Downloads",
                },
                ("13", "45", "91"): {
                    "title": "vBulletin Style Packs",
                },
                ("13", "45", "92"): {
                    "title": "vBulletin Language Packs",
                },
                ("13", "45", "93"): {
                    "title": "vBulletin Graphics and Other Resources",
                },
                ("12",): {
                    "title": "General",
                },
                ("12", "79"): {
                    "title": "vBulletin Showcase",
                },
                ("12", "32"): {
                    "title": "vBulletin Hosting Options",
                },
                ("12", "33"): {
                    "title": "PHP & HTML Questions",
                },
                ("12", "30"): {
                    "title": "Chit Chat",
                },
                ("19",): {
                    "title": "vBulletin Legacy Versions & Products",
                },
                ("19", "73"): {
                    "title": "Legacy vBulletin Versions",
                },
                ("19", "73", "100"): {
                    "title": "vBulletin 3.7 Questions, Problems and Troubleshooting",
                },
                ("19", "73", "96"): {
                    "title": "vBulletin 3.6 Questions, Problems and Troubleshooting",
                },
                ("19", "73", "94"): {
                    "title": "vBulletin 3.5 'How Do I' Questions and Troubleshooting",
                },
                ("19", "73", "86"): {
                    "title": "vBulletin 3.0 How Do I and Troubleshooting Forum",
                },
                ("19", "73", "85"): {
                    "title": "vBulletin 2 'How Do I' and Troubleshooting",
                },
                ("19", "73", "83"): {
                    "title": "vBulletin 1.1.x Archives",
                },
                ("19", "73", "84"): {
                    "title": "vBulletin Lite Archives",
                },
                ("19", "74"): {
                    "title": "Other Legacy Products",
                },
                ("19", "74", "99"): {
                    "title": "Pre-4.X Blog Archives",
                },
                ("19", "74", "98"): {
                    "title": "Project Tools Archives",
                },
                ("19", "74", "64"): {
                    "title": "vBulletin Facebook App",
                },
                ("19", "74", "44"): {
                    "title": "vBulletin Impex Import System",
                },
            },
        },
        {
            "url": "https://forum.vbulletin.com/forum/vbulletin-5-connect/vbulletin-5-connect-questions-problems-troubleshooting/vbulletin-5-tutorials/4131238-how-to-put-ads-after-first-post-in-thread-on-vb5-without-using-plugins",
            "test_base_url": "https://forum.vbulletin.com/",
            "test_contents_hash": "1f6125f1221c38cbac359e3ad2d618762ebc2f02",
            "test_item_count": 66,
        },
        {
            "url": "https://forum.vbulletin.com/forum/vbulletin-5-connect/vbulletin-5-connect-feedback/417300-cms-and-vbulletin-5",
            "test_base_url": "https://forum.vbulletin.com/",
            "test_contents_hash": "e358347798df127d65e4f96d9c95011063912ff2",
            "test_item_count": 42,
        },
    ]

    _forum_id_regex = re.compile(r"^forum(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(url)
        soup = Soup(response.content)

        generator_meta = soup.find("meta", attrs={"name": "generator"})
        if not generator_meta.get("content").startswith("vBulletin"):
            return None

        base = soup.find("base")
        return VbulletinExtractor(session, base.get("href"), options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        trs = soup.find_all("tr", class_=["category-header", "forum-item"])
        category_id = ""

        for tr in trs:
            if "category-header" in tr.get_list("class"):
                category_id = regex_match(self._forum_id_regex, tr.get("id")).group(1)

                category_anchor = tr.find("a", class_="category")
                title = category_anchor.string

                self._set_board(
                    path=(category_id,),
                    url=category_anchor.get("href"),
                    origin=response.url,
                    data={},
                    title=title,
                    are_subboards_fetched=True,
                )
            else:
                board_id = regex_match(self._forum_id_regex, tr.get("id")).group(1)

                board_anchor = tr.find("a", class_="forum-title")
                title = board_anchor.string

                self._set_board(
                    path=(category_id, board_id),
                    url=board_anchor.get("href"),
                    origin=response.url,
                    data={},
                    title=title,
                )

    def _fetch_subboards(self, board: Board):
        # Don't fetch top boards.
        if len(board.path) <= 1:
            return

        response = self._session.get(board.url)
        soup = Soup(response.content)

        trs = soup.find_all("tr", class_="forum-item")
        for tr in trs:
            subboard_id = regex_match(self._forum_id_regex, tr.get("id")).group(1)

            subboard_anchor = tr.find("a", class_="forum-title")
            self._set_board(
                path=board.path + (subboard_id,),
                url=subboard_anchor.get("href"),
                origin=response.url,
                data={},
                title=subboard_anchor.string.strip(),
            )

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        soup = Soup(response.content)

        breadcrumb_anchors = soup.find_all("a", class_="crumb-link")

        if len(breadcrumb_anchors) <= 1:
            return self.root

        # Thread.
        if soup.try_find("h2", class_="b-post__title"):
            breadcrumb_urls = [
                self._resolve_url(anchor.get("href")) for anchor in breadcrumb_anchors
            ]
            board = self.find_board_from_urls(tuple(breadcrumb_urls[2:]))

            id = soup.find("input", attrs={"name": "nodeid"}).get("value")
            title_h1 = soup.find("h1", class_="main-title")

            return Thread(
                path=board.path + (id,),
                url=urljoin(self.base_url, url),
                origin=response.url,
                data={},
                title=title_h1.string,
            )
        # Board.
        else:
            self._fetch_lower_boards(self.root)

            board_title = breadcrumb_anchors[-1].string

            for cur_board in self._boards:
                if cur_board.title == board_title:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        if not state.url:
            return None

        response = self._session.get(state.url)
        soup = Soup(response.content)

        thread_trs = soup.find_all("tr", class_="topic-item")
        for thread_tr in thread_trs:
            thread_id = thread_tr.get("data-node-id")
            thread_anchor = thread_tr.find("a", class_="topic-title")

            yield Thread(
                path=board.path + (thread_id,),
                url=thread_anchor.get("href"),
                origin=response.url,
                data={},
                title=thread_anchor.string,
            )

        next_page_anchor = soup.try_find("a", class_="right-arrow")
        if next_page_anchor and next_page_anchor.get("href"):
            return PageState(url=next_page_anchor.get("href"))

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        post_lis = soup.find_all("li", class_="b-post")
        for post_li in post_lis:
            # No support for comments for now.
            if "b-comment" in post_li.get_list("class"):
                continue

            url_anchor = post_li.find("a", class_="b-post__count")
            content_div = post_li.find("div", class_="js-post__content-text")
            author_anchor = post_li.find("div", class_="author").find("a")
            id = post_li.get("data-node-id")

            yield Post(
                path=thread.path,
                subpath=(id,),
                url=url_anchor.get("href"),
                origin=response.url,
                data={},
                author=author_anchor.string,
                content="".join(str(v) for v in content_div.contents).strip(),
            )

        next_page_anchor = soup.try_find("a", class_="right-arrow")
        if next_page_anchor and next_page_anchor.try_get("href"):
            return PageState(url=next_page_anchor.get("href"))



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from pathlib import PurePosixPath
from urllib.parse import urljoin, urlparse, urlunparse
from dataclasses import dataclass

from .common import get_relative_url, normalize_url
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


@dataclass
class DiscourseThreadPageState(PageState):
    stream_data: list[int]


class DiscourseExtractor(Extractor):
    tests = [
        {
            "url": "https://meta.discourse.org/",
            "test_base_url": "https://meta.discourse.org/",
            "test_boards": {
                ("67",): {
                    "title": "announcements",
                },
            },
        },
        {
            "url": "https://meta.discourse.org/c/announcements",
            "test_base_url": "https://meta.discourse.org/",
            "test_boards": {
                ("67", "13"): {
                    "title": "blog",
                },
            },
            "test_min_item_count": 400,
            "test_items": {
                1: {
                    "title": "About the announcements category",
                    "path": ("67", "68629"),
                }
            },
        },
        {
            "url": "https://meta.discourse.org/t/welcome-to-meta-discourse-org/1",
            "test_base_url": "https://meta.discourse.org/",
            "test_contents_hash": "36eab8dfa910b45bf8757d5d977743b90405f53f",
            "test_item_count": 1,
        },
        {
            "url": "https://meta.discourse.org/t/try-out-the-new-sidebar-and-notification-menus/238821",
            "test_base_url": "https://meta.discourse.org/",
            "test_min_item_count": 200,
        },
    ]

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        url = url.removesuffix("/").removesuffix(".json")
        url = url.removesuffix(".json")

        response = session.get_noretry(normalize_url(url))
        soup = Soup(response.content)

        crawler_nav = soup.find("nav", class_="crawler-nav")
        home_anchor = crawler_nav.find("a")
        base_url = urljoin(response.url, home_anchor.get("href"))

        return DiscourseExtractor(session, normalize_url(base_url), options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True
        response = self._session.get(urljoin(self.base_url, "site.json"))
        site_json = response.json()

        for category_data in site_json["categories"]:
            if "parent_category_id" not in category_data:
                id = str(category_data["id"])

                self._set_board(
                    path=(id,),
                    url=urljoin(self.base_url, f"c/{category_data['slug']}/{id}"),
                    origin=response.url,
                    data=category_data,
                    title=category_data["name"],
                    are_subboards_fetched=True,
                )

        for category_data in site_json["categories"]:
            if "parent_category_id" in category_data:
                slug = category_data["slug"]
                id = str(category_data["id"])
                parent_id = str(category_data["parent_category_id"])

                self._set_board(
                    path=(parent_id, id),
                    url=urljoin(self.base_url, f"c/{slug}/{id}"),
                    origin=response.url,
                    data=category_data,
                    title=category_data["name"],
                    are_subboards_fetched=True,
                )

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        url = url.removesuffix(".json")

        relative_url = get_relative_url(url, self.base_url)
        url_parts = PurePosixPath(relative_url).parts

        if len(url_parts) <= 1:
            return self.root

        if url_parts[0] == "c":
            slug = url_parts[1]

            for _, board in self._subboards[self.root.path].items():
                if board.data["slug"] == slug:
                    return board

                for _, subboard in self._subboards[board.path].items():
                    if subboard.data["slug"] == slug:
                        return subboard
        elif url_parts[0] == "t":
            id = url_parts[1]
            json_url = urljoin(self.base_url, f"t/{id}.json")
            response = self._session.get(json_url)
            data = response.json()

            slug = data["slug"]
            category_id = str(data["category_id"])

            if category_id in self._subboards[self.root.path]:
                path = (category_id, f"{id}")
            else:
                for _, subboard in self._subboards[self.root.path].items():
                    if category_id in self._subboards[self.root.path]:
                        path = subboard.path + (category_id, f"{id}")
                        break
                else:
                    raise ValueError

            return Thread(
                path=path,
                url=url,
                origin=response.url,
                data=data,
                title=data["title"],
            )

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if state.url == board.url:
            relative_url = get_relative_url(state.url, self.base_url)
            url_parts = PurePosixPath(relative_url).parts

            if len(url_parts) <= 1 or url_parts[0] != "c":
                return None

            state.url = f"{state.url}.json"

        response = self._session.get(state.url)
        page_json = response.json()

        for data in page_json["topic_list"]["topics"]:
            id = str(data["id"])
            yield Thread(
                path=board.path + (id,),
                url=urljoin(self.base_url, f"t/{data['slug']}/{id}"),
                origin=response.url,
                data=data,
                title=data["title"],
            )

        if more_topics_url := page_json["topic_list"].get("more_topics_url", None):
            parsed_more_topics_url = urlparse(str(more_topics_url))
            parsed_more_topics_url = parsed_more_topics_url._replace(
                path=f"{parsed_more_topics_url.path}.json"
            )

            return PageState(
                url=urljoin(self.base_url, str(urlunparse(parsed_more_topics_url)))
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        if state.url == thread.url:
            json_url = f"{state.url}.json"
            response = self._session.get(json_url)
            page_json = response.json()
            state = DiscourseThreadPageState(
                url=response.url, stream_data=page_json["post_stream"]["stream"]
            )
        else:
            origin = state.url
            state = cast(DiscourseThreadPageState, state)
            post_ids = tuple(state.stream_data[:20])
            response = self._session.uncached_get(
                origin, params={"post_ids[]": post_ids}
            )
            page_json = response.json()

        datas = page_json["post_stream"]["posts"]

        for data in datas:
            topic_slug = data["topic_slug"]

            state.stream_data.pop(0)
            yield Post(
                path=thread.path,
                subpath=(str(data["id"]),),
                url=urljoin(self.base_url, f"t/{topic_slug}/{id}"),
                origin=response.url,
                data=data,
                author=data.get("username", None),
                content=data.get("cooked", None),
            )

        topic_id = str(page_json["id"])

        if state.stream_data:
            state.url = urljoin(self.base_url, f"t/{topic_id}/posts.json")
            return state



# type: ignore
from __future__ import annotations
from typing import *  # type: ignore

from dataclasses import dataclass
from functools import lru_cache, wraps
from tenacity import (
    retry,
    wait_random_exponential,
    stop_after_attempt,
    before_sleep_log,
)
import requests
import time
import logging

from .exceptions import SearchError
from .version import __version__


def hash_dict(func):
    class hashdict(dict):
        def __hash__(self):
            return hash(frozenset(self.items()))

    @wraps(func)
    def wrapped(*args, **kwargs):
        args = tuple([hashdict(arg) if isinstance(arg, dict) else arg for arg in args])
        kwargs = {
            k: hashdict(v) if isinstance(v, dict) else v for k, v in kwargs.items()
        }
        return func(*args, **kwargs)

    return wrapped


@dataclass  # (kw_only=True)
class SessionOptions:
    user_agent: str
    get_urls: bool


class Session:
    def __init__(self, options: SessionOptions):
        self._session = requests.Session()
        self._options = options
        self.delay = 1
        self.attempts = 0

    @hash_dict
    @lru_cache(maxsize=1024)
    def get(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        if self._options.get_urls:
            print(url)
        else:
            logging.info(f"GET {url}")

        return self._get(url, params, headers, **kwargs)

    def uncached_get(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        if self._options.get_urls:
            print(url)
        else:
            logging.info(f"GET (uncached) {url}")

        return self._get(url, params, headers, **kwargs)

    def get_noretry(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        if not headers:
            headers = {"User-Agent": self._options.user_agent}

        response = self._session.get(url, params=params, headers=headers, **kwargs)

        if response.status_code != 200:
            raise SearchError

        return response

    def try_get(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        if not headers:
            headers = {"User-Agent": self._options.user_agent}

        return self._session.get(url, params=params, headers=headers, **kwargs)

    def _after_retry(self):
        logging.warning(f"Waiting {self.delay} seconds.")

    @retry(
        reraise=True,
        wait=wait_random_exponential(multiplier=5),
        stop=stop_after_attempt(5),
        before_sleep=before_sleep_log(logging.getLogger(), logging.WARNING),
    )
    def _get(
        self,
        url: str,
        params: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        if not headers:
            headers = {"User-Agent": self._options.user_agent}

        response = self.try_get(url, params=params, headers=headers, **kwargs)
        response.raise_for_status()

        return response



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

import dataclasses
import json

from .common import FileWriter, Entry


class JsonlWriter(FileWriter):
    def _serialize_entry(self, entry: Entry):
        return json.dumps(dataclasses.asdict(entry))



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from abc import ABC, abstractmethod
from dataclasses import dataclass
from mailbox import Mailbox, Message

try:
    from html2text import html2text
except ImportError:
    pass

from datetime import datetime, timezone
import sys

from ..extractors.common import Extractor, Item, Thread, Board, Post, PageState
from ..version import __version__


@dataclass  # (kw_only=True)
class WriterOptions:
    output_path: str
    write_board_objects: bool
    write_thread_objects: bool
    write_post_objects: bool
    content_as_title: bool
    textify: bool


@dataclass  # (kw_only=True)
class WriterState:
    board_path: tuple[str, ...] | None = None
    board_page: PageState | None = None
    thread_page: PageState | None = None


@dataclass  # (kw_only=True)
class Entry:
    generator: str
    version: str
    extractor: str
    download_time: str
    type: str
    item: Item


class Writer(ABC):
    tests: list[dict[str, Any]]

    def __init__(self, extractor: Extractor, options: WriterOptions):
        self._extractor = extractor
        self._options = options
        self._initial_state = WriterState()

    def write(self, url: str):
        self.read_metadata()

        base_node = self._extractor.node_from_url(url)

        if isinstance(base_node, Board):
            self.write_board(base_node)
        elif isinstance(base_node, Thread):
            self.write_thread(base_node)

    @abstractmethod
    def read_metadata(self):
        pass

    @abstractmethod
    def _write_board_object(self, board: Board):
        pass

    def _write_board_threads(self, board: Board):
        for thread in self._extractor.threads(board, self._initial_state.board_page):
            self.write_thread(thread)

    @final
    def write_board(self, board: Board):
        if self._options.write_board_objects:
            self._write_board_object(board)

        self._write_board_threads(board)

        for _, subboard in self._extractor.subboards(board).items():
            self.write_board(subboard)

    @abstractmethod
    def _write_thread_object(self, thread: Thread):
        pass

    def _write_thread_posts(self, thread: Thread):
        for post in self._extractor.posts(thread, self._initial_state.thread_page):
            self.write_post(thread, post)

    @final
    def write_thread(self, thread: Thread):
        if self._options.write_thread_objects:
            self._write_thread_object(thread)

        self._write_thread_posts(thread)

    @abstractmethod
    def _write_post_object(self, thread: Thread, post: Post):
        pass

    @final
    def write_post(self, thread: Thread, post: Post):
        if self._options.write_post_objects:
            self._write_post_object(thread, post)


class SimulatedWriter(Writer):
    def read_metadata(self):
        pass

    def _write_board_object(self, board: Board):
        pass

    def _write_thread_object(self, thread: Thread):
        pass

    def _write_post_object(self, thread: Thread, post: Post):
        pass


class FileWriter(Writer):
    def __init__(self, extractor: Extractor, options: WriterOptions):
        super().__init__(extractor, options)

        if options.output_path != "-":
            self._file = open(options.output_path, "w")
        else:
            self._file = None

    def __del__(self):
        if self._file:
            self._file.close()

    def read_metadata(self):
        pass  # TODO.

    def _write_board_object(self, board: Board):
        entry = self._make_entry(board)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _write_thread_object(self, thread: Thread):
        entry = self._make_entry(thread)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _write_post_object(self, thread: Thread, post: Post):
        entry = self._make_entry(post)

        if self._file:
            self._file.write(f"{self._serialize_entry(entry)}\n")
        else:
            sys.stdout.write(f"{self._serialize_entry(entry)}\n")

    def _make_entry(self, item: Item):
        if isinstance(item, Board):
            type = "board"
        elif isinstance(item, Thread):
            type = "thread"
        elif isinstance(item, Post):
            type = "post"
        else:
            raise ValueError

        return Entry(
            generator="forum-dl",
            version=__version__,
            extractor=self._extractor.__class__.__module__.split(".")[-1],
            download_time=datetime.now(timezone.utc).isoformat(),
            type=type,
            item=item,
        )

    @abstractmethod
    def _serialize_entry(self, entry: Entry) -> str:
        pass


class MailWriter(Writer):
    def __init__(
        self,
        extractor: Extractor,
        mailbox: Mailbox[Any],
        options: WriterOptions,
    ):
        super().__init__(extractor, options)
        self._mailbox = mailbox

        for key, msg in self._mailbox.iteritems():
            if msg.get("X-Forumdl-Version"):
                self._metadata_key = key
        else:
            msg = self._new_message()
            self._metadata_key = self._mailbox.add(msg)

    def __del__(self):
        self._mailbox.flush()
        self._mailbox.close()

    def write(self, url: str):
        self._mailbox.lock()
        super().write(url)
        self._mailbox.unlock()

    def read_metadata(self):
        pass  # TODO.

    def _write_board_object(self, board: Board):
        pass  # TODO.

    def _write_thread_object(self, thread: Thread):
        pass  # TODO.

    def _write_post_object(self, thread: Thread, post: Post):
        self._mailbox.add(self._build_message(thread, post))

    @abstractmethod
    def _new_message(self) -> Message:
        pass

    def _build_message(self, thread: Thread, post: Post):
        msg = self._new_message()

        path = post.path + post.subpath

        msg["Message-ID"] = "<" + ".".join(path) + ">"
        msg["From"] = post.author

        if len(path) >= 2:
            msg["In-Reply-To"] = f"<{'.'.join(path[:-1])}>"

            refs = f"{path[0]}"
            for ref in post.path[1:-1]:
                refs += f" <{ref}>"

        if len(post.subpath) >= 1 and self._options.content_as_title:
            msg["Subject"] = html2text(post.content[:98]).partition("\n")[0]
        else:
            msg["Subject"] = thread.title

        # msg["Date"] = formatdate(post.date)

        if self._options.textify:
            msg.set_type("text/plain")
            msg.set_payload(html2text(post.content), "utf-8")
        else:
            msg.set_type("text/html")
            msg.set_payload(post.content, "utf-8")

        return msg


class FolderedMailWriter(MailWriter):
    def __init__(
        self,
        extractor: Extractor,
        mailbox: Mailbox[Any],
        options: WriterOptions,
    ):
        super().__init__(extractor, mailbox, options)
        self.folders: dict[str, Mailbox[Any]] = {}

    def _folder_name(self, board: Board):
        return ".".join(board.path)

    def _write_board_object(self, board: Board):
        folder_name = self._folder_name(board)
        self.folders[folder_name] = getattr(self._mailbox, "add_folder")(folder_name)
        super()._write_board_object(board)

    def _write_post_object(self, thread: Thread, post: Post):
        board = self._extractor.find_board(thread.path[:-1])
        folder_name = self._folder_name(board)
        self.folders[folder_name].add(self._build_message(thread, post))



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

import argparse
import logging

from .version import __version__


def build_parser():
    parser = argparse.ArgumentParser(
        add_help=False,
    )

    general = parser.add_argument_group("General Options")
    general.add_argument(
        "--help",
        action="help",
        help="Show this help message and exit",
    )
    general.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print program version and exit",
    )
    general.add_argument(
        "--list-extractors",
        dest="list_extractors",
        action="store_true",
        help="List all supported extractors and exit",
    )
    general.add_argument(
        "--list-output-formats",
        dest="list_output_formats",
        action="store_true",
        help="List all supported output formats and exit",
    )

    general.add_argument(
        "--user-agent",
        dest="user_agent",
        default=f"Forum-dl {__version__}",
        help="User-Agent request header",
    )

    output = parser.add_argument_group("Output Options")
    output.add_argument(
        "-q",
        "--quiet",
        dest="loglevel",
        default=logging.INFO,
        action="store_const",
        const=logging.ERROR,
        help="Activate quiet mode",
    )
    output.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        action="store_const",
        const=logging.DEBUG,
        help="Print various debugging information",
    )
    output.add_argument(
        "-g",
        "--get-urls",
        dest="get_urls",
        action="store_true",
        help="Print URLs instead of downloading",
    )
    output.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        dest="output",
        help="Output all results concatenated to FILE, or stdout if FILE is -",
    )
    output.add_argument(
        "-f",
        "--output-format",
        metavar="FORMAT",
        dest="output_format",
        default="jsonl",
        help="Output format. Use --list-output-formats for a list of possible arguments",
    )
    output.add_argument(
        "--no-boards",
        dest="boards",
        action="store_false",
        help="Do not write board objects",
    )
    output.add_argument(
        "--no-threads",
        dest="threads",
        action="store_false",
        help="Do not write thread objects",
    )
    output.add_argument(
        "--no-posts",
        dest="posts",
        action="store_false",
        help="Do not write post objects",
    )
    output.add_argument(
        "--textify",
        dest="textify",
        action="store_true",
        help="Lossily convert HTML content to plaintext",
    )
    output.add_argument(
        "--content-as-title",
        dest="content_as_title",
        action="store_true",
        help="Write 98 initial characters of content in title field of each post",
    )

    parser.add_argument(
        "urls",
        metavar="URL",
        nargs="*",
        help=argparse.SUPPRESS,
    )

    return parser



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from urllib.parse import urljoin
import re

from .common import normalize_url, regex_match
from .common import Extractor, ExtractorOptions, Board, Thread, Post, PageState
from ..session import Session
from ..soup import Soup


class XenforoExtractor(Extractor):
    tests = [
        {
            "url": "https://xenforo.com/community",
            "test_base_url": "https://xenforo.com/community/",
            "test_boards": {
                ("1",): {
                    "title": "Official forums",
                },
                ("1", "2"): {
                    "title": "Announcements",
                },
                ("1", "3"): {
                    "title": "Have you seen...?",
                },
                ("1", "19"): {
                    "title": "Frequently asked questions",
                },
                ("17",): {
                    "title": "Public forums",
                },
                ("17", "5"): {
                    "title": "XenForo pre-sales questions",
                },
                ("17", "18"): {
                    "title": "XenForo suggestions",
                },
                ("46",): {
                    "title": "XenForo bug reports",
                },
                ("46", "91"): {
                    "title": "Bug reports",
                },
                ("46", "105"): {
                    "title": "Importer bug reports",
                },
                ("46", "92"): {
                    "title": "Resolved bug reports",
                },
                ("46", "43"): {
                    "title": "Future-fix bug reports",
                },
                ("22",): {
                    "title": "Customer forums",
                },
                ("22", "23"): {
                    "title": "Installation, upgrade, and import support",
                },
                ("22", "24"): {
                    "title": "Troubleshooting and problems",
                },
                ("22", "25"): {
                    "title": "XenForo questions and support",
                },
                ("22", "47"): {
                    "title": "Styling and customization questions",
                },
                ("22", "48"): {
                    "title": "Server configuration and hosting",
                },
                ("22", "84"): {
                    "title": "Forum management",
                },
                ("22", "79"): {
                    "title": "Tips and guides [1.x]",
                },
                ("22", "94"): {
                    "title": "Tips and guides [2.x]",
                },
                ("22", "45"): {
                    "title": "Other XenForo discussions and feedback",
                },
                ("22", "49"): {
                    "title": "XenForo help and manual",
                },
                ("54",): {
                    "title": "Official XenForo add-ons",
                },
                ("54", "86"): {
                    "title": "Media Gallery suggestions",
                },
                ("54", "87"): {
                    "title": "Media Gallery support",
                },
                ("54", "88"): {
                    "title": "Media Gallery bug reports",
                },
                ("54", "65"): {
                    "title": "Resource Manager suggestions",
                },
                ("54", "82"): {
                    "title": "Resource Manager support",
                },
                ("54", "66"): {
                    "title": "Resource Manager bug reports",
                },
                ("54", "55"): {
                    "title": "Enhanced Search suggestions",
                },
                ("54", "56"): {
                    "title": "Enhanced Search support",
                },
                ("54", "57"): {
                    "title": "Enhanced Search bug reports",
                },
                ("60",): {
                    "title": "XenForo resources and add-ons",
                },
                ("60", "70"): {
                    "title": "Resources and add-ons",
                },
                ("60", "71"): {
                    "title": "Resource and add-on releases [1.x]",
                },
                ("60", "93"): {
                    "title": "Resource and add-on releases [2.x]",
                },
                ("60", "68"): {
                    "title": "Resource and add-on requests",
                },
                ("60", "63"): {
                    "title": "Resource and add-on discussions",
                },
                ("60", "69"): {
                    "title": "Custom service/development requests",
                },
                ("60", "42"): {
                    "title": "Third-party services & offers",
                },
                ("60", "62"): {
                    "title": "Resource and add-on archive",
                },
                ("51",): {
                    "title": "Development help",
                },
                ("51", "106"): {
                    "title": "Resource standards",
                },
                ("51", "80"): {
                    "title": "Development tutorials and resources [1.x]",
                },
                ("51", "101"): {
                    "title": "Development tutorials [2.x]",
                },
                ("51", "34"): {
                    "title": "XenForo development discussions",
                },
                ("51", "52"): {
                    "title": "General PHP and MySQL discussions",
                },
                ("4",): {
                    "title": "General discussions",
                },
                ("4", "53"): {
                    "title": "Forum showcase and critiques",
                },
                ("4", "7"): {
                    "title": "Off topic",
                },
                ("14",): {
                    "title": "Tests and examples",
                },
                ("14", "109"): {
                    "title": "Resolved bugs",
                },
                ("14", "110"): {
                    "title": "Implemented suggestions",
                },
                ("14", "11"): {
                    "title": "Example page",
                },
            },
        },
        {
            "url": "https://xenforo.com/community/threads/darkness-free-xenforo-2-gaming-skin-deleted.154782/",
            "test_base_url": "https://xenforo.com/community/",
            "test_contents_hash": "00cd24c8a7aba0f567d6a0868cfe27fada450ca4",
            "test_item_count": 53,
        },
        {
            "url": "https://xenforo.com/community/threads/steam-authentication-deleted.141309/",
            "test_base_url": "https://xenforo.com/community/",
            "test_contents_hash": "47091f2cb3f716ee16f5057f8350a2f514ed4bce",
            "test_item_count": 26,
        },
    ]

    _category_class_regex = re.compile(r"^block--category(\d+)$")
    _board_class_regex = re.compile(r"^node--id(\d+)$")
    _thread_class_regex = re.compile(r"^js-threadListItem-(\d+)$")
    _thread_key_regex = re.compile(r"^thread-(\d+)$")
    _post_id_regex = re.compile(r"^post-(\d+)$")

    @staticmethod
    def _detect(session: Session, url: str, options: ExtractorOptions):
        response = session.get_noretry(
            normalize_url(url, remove_suffixes=[], append_slash=False)
        )
        soup = Soup(response.content)

        data_nav_id_anchor = soup.find("a", attrs={"data-nav-id": "forums"})
        base_url = normalize_url(urljoin(url, data_nav_id_anchor.get("href")))
        if not base_url:
            return None

        if not soup.find("a", attrs={"rel": "sponsored noopener"}):
            return None

        return XenforoExtractor(session, base_url, options)

    def _fetch_top_boards(self):
        self._are_subboards_fetched[self.root.path] = True

        response = self._session.get(self.base_url)
        soup = Soup(response.content)

        block_category_divs = soup.find_all("div", class_=self._category_class_regex)
        for block_category_div in block_category_divs:
            category_header = block_category_div.find("h2", class_="block-header")
            category_anchor = category_header.find("a")
            category_id = regex_match(
                self._category_class_regex, block_category_div.get_list("class")
            ).group(1)
            category_href = category_header.find("a").get("href")

            self._set_board(
                path=(category_id,),
                url=urljoin(response.url, category_href),
                origin=response.url,
                data={},
                title=category_anchor.string.strip(),
                are_subboards_fetched=True,
            )

            node_id_divs = block_category_div.find_all(
                "div", class_=self._board_class_regex
            )

            for node_id_div in node_id_divs:
                subboard_id = regex_match(
                    self._board_class_regex, node_id_div.get_list("class")
                ).group(1)

                node_description_anchor = node_id_div.find(
                    "a", attrs={"data-shortcut": "node-description"}
                )

                href = node_description_anchor.get("href")

                self._set_board(
                    path=(category_id, subboard_id),
                    url=urljoin(self.base_url, href),
                    origin=response.url,
                    data={},
                    title=node_description_anchor.string.strip(),
                )

        self._fetch_lower_boards(self.root)

    def _fetch_subboards(self, board: Board):
        pass

    def _get_node_from_url(self, url: str):
        response = self._session.get(url)
        soup = Soup(response.content)

        breadcrumbs_ul = soup.find("ul", class_="p-breadcrumbs")
        breadcrumb_anchors = breadcrumbs_ul.find_all("a", attrs={"itemprop": "item"})

        if len(breadcrumb_anchors) <= 1:
            return self.root

        # Thread.
        if soup.find("article"):
            board_url = urljoin(url, breadcrumb_anchors[-2].get("href"))
            html = soup.find("html")
            thread_id = regex_match(
                self._thread_key_regex, html.get("data-content-key")
            ).group(1)
            title_h1 = soup.find("h1", class_="p-title-value")

            for cur_board in self._boards:
                if cur_board.url == board_url:
                    return Thread(
                        path=cur_board.path + (thread_id,),
                        url=urljoin(self.base_url, url),
                        origin=response.url,
                        data={},
                        title=title_h1.string,
                    )
        # Board.
        else:
            board_href = self._resolve_url(breadcrumb_anchors[-1].get("href"))

            for cur_board in self._boards:
                if cur_board.url == board_href:
                    return cur_board

        raise ValueError

    def _fetch_lazy_subboard(self, board: Board, id: str):
        pass

    def _fetch_lazy_subboards(self, board: Board):
        yield from ()

    def _fetch_board_page_threads(self, board: Board, state: PageState):
        if board == self.root:
            return None

        if not state.url:
            return None

        response = self._session.get(state.url)
        soup = Soup(response.content)

        thread_divs = soup.find_all("div", class_=self._thread_class_regex)
        for thread_div in thread_divs:
            thread_id = regex_match(
                self._thread_class_regex, thread_div.get_list("class")[-1]
            ).group(1)

            title_div = thread_div.find("div", class_="structItem-title")
            title_anchor = title_div.find("a")

            url = urljoin(self.base_url, title_anchor.get("href"))

            yield Thread(
                path=board.path + (thread_id,),
                url=url,
                origin=response.url,
                data={},
                title=title_anchor.string,
            )

        next_page_anchor = soup.try_find("a", class_="pageNav-jump--next")
        if next_page_anchor:
            return PageState(
                url=urljoin(self.base_url, next_page_anchor.get("href")),
            )

    def _fetch_thread_page_posts(self, thread: Thread, state: PageState):
        response = self._session.get(state.url)
        soup = Soup(response.content)

        message_articles = soup.find_all("article", class_="message")

        for message_article in message_articles:
            bbwrapper_div = message_article.find("div", class_="bbWrapper")
            message_attribution_ul = message_article.find(
                "ul", class_="message-attribution-main"
            )
            url_anchor = message_attribution_ul.find("a")

            yield Post(
                path=thread.path,
                subpath=(
                    regex_match(
                        self._post_id_regex, message_article.get("data-content")
                    ).group(1),
                ),
                url=urljoin(state.url, url_anchor.get("href")),
                origin=response.url,
                data={},
                author=message_article.get("data-author"),
                content=str(bbwrapper_div.encode_contents()),
            )

        next_page_anchor = soup.try_find("a", class_="pageNav-jump--next")
        if next_page_anchor:
            return PageState(
                url=urljoin(self.base_url, next_page_anchor.get("href")),
            )



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore


class ForumDlException(Exception):
    pass


class ExtractorNotFoundError(ForumDlException):
    pass


class WriterNotFoundError(ForumDlException):
    pass


class SearchError(ForumDlException):
    pass


class TagSearchError(SearchError):
    pass


class AttributeSearchError(SearchError):
    pass


class PropertyError(SearchError):
    pass



# pyright: strict
from __future__ import annotations
from typing import *  # type: ignore

from mailbox import MMDF, MMDFMessage

from .common import FolderedMailWriter, WriterOptions
from ..extractors.common import Extractor


class MmdfWriter(FolderedMailWriter):
    tests = []

    def __init__(self, extractor: Extractor, options: WriterOptions):
        super().__init__(extractor, MMDF(options.output_path), options)

    def _new_message(self):
        return MMDFMessage()



import logging
import sys

from .forumdl import ForumDl
from . import options

from .session import SessionOptions
from .extractors.common import ExtractorOptions
from .writers.common import WriterOptions


def main():
    parser = options.build_parser()
    args = parser.parse_args()

    logging.basicConfig()
    logging.getLogger().setLevel(args.loglevel)

    forumdl = ForumDl()

    if args.list_extractors:
        print("\n".join(forumdl.list_extractors()))
    elif args.list_output_formats:
        print("\n".join(forumdl.list_output_formats()))
    elif not args.urls:
        parser.error(
            "The following arguments are required: URL\n"
            "Use 'forum-dl --help' to get a list of all options."
        )
    else:
        forumdl.download(
            urls=args.urls,
            output_format=args.output_format,
            session_options=SessionOptions(
                user_agent=args.user_agent,
                get_urls=args.get_urls,
            ),
            extractor_options=ExtractorOptions(
                path=False,
            ),
            writer_options=WriterOptions(
                output_path=args.output,
                write_board_objects=args.boards,
                write_thread_objects=args.threads,
                write_post_objects=args.posts,
                content_as_title=args.content_as_title,
                textify=args.textify,
            ),
        )
